#!/usr/bin/env python3
"""
 DESCRIPTION
          This tool is part of cygnsslib python package. The package is created by Mixil lab at USC
          See <https://bitbucket.org/usc_mixil/cygnss-library>

          This Tool download CYGNSS data

 AUTHOR   Amer Melebari
          Microwave Systems, Sensors and Imaging Lab (MiXiL)
          University of Southern California
 EMAIL    amelebar@usc.edu
 CREATED  2020‑07‑19
 Updated  2023-05-02

  Copyright 2023 University of Southern California
"""
import logging
from inspect import signature
from typing import Optional, Union
from cygnsslib.data_downloader.download_srtm import EarthdataSession
import subscriber.podaac_access as pa
from time import sleep
from tqdm.auto import tqdm
import datetime as dt
import fnmatch
import hashlib
import numpy as np
import os
import requests
import shutil
import warnings
from urllib.error import HTTPError


CYG_MIN_FILE_SIZE = 50e6  # in bytes
PODAAC_CYG_URL = 'https://podaac-tools.jpl.nasa.gov/drive/files/allData/cygnss'
OPEN_DAP_URL = 'https://opendap.jpl.nasa.gov/opendap/allData/cygnss'
L1_VER = 'v3.2'
CHUNK_SIZE = 1024 * 1024  # 1 MB
MAX_CONNECTIONS_TRIES = 10
page_size = 2000


def cyg_ver2shortname(cyg_ver: str, cyg_data_level: str):
    if 'raw' in cyg_data_level.lower():
        return 'CYGNSS_L1_RAW_IF'
    else:
        return f'CYGNSS_{cyg_data_level.upper():s}_{cyg_ver.upper():s}'


def checksum(file_path, chunk_num_blocks=4096):
    """
    do md5 checksum for large files

    :param file_path: file path
    :type file_path: str
    :param chunk_num_blocks: number of blocks in a chunk
    :type chunk_num_blocks: int
    :return:
    """

    h = hashlib.md5()
    with open(file_path, 'rb') as f:
        while chunk := f.read(chunk_num_blocks * h.block_size):
            h.update(chunk)
    return h.hexdigest()


def download_file(file_url: str, output_folder: str, auth: Optional[tuple[str, str]] = None, url_md5_checksum: Optional[str] = None) -> Optional[str]:
    """
    download the file with url into folder output_folder

    :param file_url: url of the file
    :type file_url: str
    :param output_folder: saving folder
    :type output_folder: str
    :param auth: username or pass
    :type auth: tuple of str or None
    :param url_md5_checksum: md5 checksum of the downloaded file, if None, the code will find it
    :type url_md5_checksum: str
    :return: downloaded file path
    :rtype: str
    """
    num_redownload = 0
    file_name = file_url.split('/')[-1]
    out_file = os.path.join(output_folder, file_name)
    out_file_temp = f'{out_file:s}.incomplete'
    for i_try in range(MAX_CONNECTIONS_TRIES):
        try:
            with EarthdataSession(username=auth[0], password=auth[1]) as session:
                with session.get(file_url, stream=True) as response:
                    if response.status_code == 404:
                        return None
                    elif response.status_code == 401:  # Auth Error
                        raise requests.exceptions.RequestException
                    response.raise_for_status()
                    response.raw.decode_content = True
                    file_size = int(response.headers.get('content-length', 0))
                    free_disk_space = shutil.disk_usage(output_folder).free
                    if file_size > free_disk_space:
                        def bytes2mb(x: float) -> int: return int(x / 1024 / 1024)
                        raise IOError(f'No enough space in the disk. file size: {bytes2mb(file_size):d} MB, free space {bytes2mb(free_disk_space):d}')
                    file_screen_name = f'...{os.sep.join(out_file_temp.split(os.sep)[-4:])[:-11]}' if (len(out_file_temp) > 80) else out_file_temp
                    with tqdm.wrapattr(open(out_file_temp, "wb"), "write", miniters=1, total=file_size, desc=file_screen_name) as target_file:
                        for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                            target_file.write(chunk)
                if url_md5_checksum is None:  # get md5 checksum
                    url_md5_checksum = session.get(f'{file_url:s}.md5').content.strip().decode('utf').split()[0]
                if url_md5_checksum:  # if the md5 file available
                    file_checksum = checksum(out_file_temp)
                    if url_md5_checksum != file_checksum:
                        raise RuntimeError
                else:
                    print(f'MD5 checksum is not available for this file; {file_name:s}')
        except RuntimeError:  # MD5 verification failed
            if i_try < MAX_CONNECTIONS_TRIES and num_redownload < 2:
                print("MD5 checksum failed! Trying to re-download it again")
                os.remove(out_file_temp)
                num_redownload += 1
            else:
                print("MD5 checksum failed!")
                shutil.move(out_file_temp, out_file)
                break
        except (requests.exceptions.ConnectionError, requests.exceptions.HTTPError):
            if i_try < MAX_CONNECTIONS_TRIES:
                print('Connection Error sleep for 5 sec')
                sleep(5)
            else:
                raise RuntimeError(f'a connection error was raised when trying to access {file_url:s}')
        else:  # if it was successful
            shutil.move(out_file_temp, out_file)
            break
    return out_file


def download_cyg_files(data_year: int, list_data_day: Union[list[int], int], list_sc_num: Optional[Union[list[int], int]] = None,
                       cyg_data_ver: Optional[str] = None, cyg_data_lvl: str = 'L1', cygnss_l1_path: Optional[str] = None,
                       checksum_exist_file: bool = False, force_download: bool = False):
    """
    
    download multiple CYGNSS files
    
    :param data_year: data year
    :type data_year: int
    :param list_data_day: list of data days
    :type list_data_day: list or int or np.array
    :param list_sc_num: list of cygnss spacecraft numbers (1-8), if None will download all SCs
    :type list_sc_num: list or int or np.array or None
    :param cyg_data_ver: cygnss data version (ex: 'v2.1')
    :type cyg_data_ver: str
    :param cyg_data_lvl: cygnss data level (ex: 'L1')
    :type cyg_data_lvl: str
    :param cygnss_l1_path: path of the cygnss L1 data (default: os.environ.get('CYGNSS_L1_PATH'))
    :type cygnss_l1_path: str or None
    :param checksum_exist_file: check md5 checksum for existing files? this will make it very slow
    :type checksum_exist_file: bool
    :param force_download: re-download the file even if the version is not included in the path (not recommended)
    :type force_download: bool
    :return:
    """
    if cygnss_l1_path is None:
        cygnss_l1_path = os.environ.get('CYGNSS_L1_PATH')
    if np.size(list_data_day) == 1:
        list_data_day = [int(list_data_day)]
    if list_sc_num is None:
        list_sc_num = np.arange(1, 9)
    elif np.size(list_sc_num) == 1:
        list_sc_num = [int(list_sc_num)]
    search_keywords = [f'cyg{sc_num:02d}' for sc_num in list_sc_num]
    if cygnss_l1_path is None:
        raise ValueError("$CYGNSS_L1_PATH environment variable need to be set, or use cygnss_l1_path input parameter")

    if cyg_data_ver is None:
        folder_list = cygnss_l1_path.split(os.path.sep)
        if not folder_list[-1]:
            folder_list.pop(-1)
        cyg_data_ver = folder_list[-1]
    cyg_data_lvl = cyg_data_lvl.upper()

    # get temporal limit for the search
    start_date_time = dt.datetime(data_year, 1, 1) + dt.timedelta(days=min(list_data_day) - 1)
    end_date_time = dt.datetime(data_year, 1, 1, 23, 59, 59) + dt.timedelta(days=max(list_data_day) - 1)
    temporal_range = pa.get_temporal_range(start_date_time.isoformat() + 'Z', end_date_time.isoformat() + 'Z',
                                           dt.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))

    # Get CYGNSS Shortname
    cyg_data_shortname = cyg_ver2shortname(cyg_data_ver, cyg_data_lvl)

    pa.setup_earthdata_login_auth(pa.edl)
    token = pa.get_token(pa.token_url)

    params = [
        ('page_size', page_size),
        ('sort_key', "start_date"),
        ('provider', 'POCLOUD'),
        ('ShortName', cyg_data_shortname),
        ('temporal', temporal_range),
        ('token', token),
    ]

    # If 401 is raised, refresh token and try one more time
    try:
        results = pa.get_search_results(params)
    except HTTPError as e:
        if e.code == 401:
            token = pa.refresh_token(token)
            # Updated: This is not always a dictionary...
            # in fact, here it's always a list of tuples
            for i, p in enumerate(params):
                if p[0] == "token":
                    params[i] = ("token", token)
            results = pa.get_search_results(params)
        else:
            raise e

    checksums = pa.extract_checksums(results)
    file_start_times = pa.parse_start_times(results)

    downloads = []
    for r in results['items']:
        for u in r['umm']['RelatedUrls']:
            if u['Type'] == "GET DATA" and ('Subtype' not in u or u['Subtype'] != "OPENDAP DATA"):
                f = u['URL']
                data_date_str = r['umm']['TemporalExtent']['RangeDateTime']['BeginningDateTime']
                data_date = dt.datetime.fromisoformat(data_date_str[:-1]).date()
                doy = data_date.timetuple().tm_yday
                if doy not in list_data_day:
                    continue
                for extension in ["\\.nc"]:
                    if pa.search_extension(extension, f):
                        for keyword in search_keywords:
                            if keyword in f.lower():
                                downloads.append(f)
                                break
    ver_folder_exist = check_ver_folder(cygnss_l1_path, cyg_data_ver)
    if not force_download and not ver_folder_exist:
        error_str = f"You are trying to download version {cyg_data_ver:s}, but the path doesn't contain the version name"
        raise ValueError(f'{error_str:s}\nuse force_download=True to remove this error')

    success_cnt = failure_cnt = skip_cnt = 0
    downloaded_files_list = []
    for f in downloads:
        try:
            output_path = prepare_time_output(file_start_times, cygnss_l1_path, f)

            # decide if we should actually download this file (e.g. we may already have the latest version)
            if os.path.exists(output_path) and not force_download and (not checksum_exist_file or pa.checksum_does_match(output_path, checksums)):
                logging.info(str(dt.datetime.now()) + " SKIPPED: " + f)
                skip_cnt += 1
                continue
            if len(list(signature(pa.download_file).parameters)) > 3:  # allow for the use of non MiXIL version
                checksum_val = checksums.get(os.path.basename(output_path))
                pa.download_file(f, output_path, checksum_val)
            else:
                pa.download_file(f, output_path)
            logging.info(str(dt.datetime.now()) + " SUCCESS: " + f)
            success_cnt += 1

        except Exception:
            logging.warning(str(dt.datetime.now()) + " FAILURE: " + f, exc_info=True)
            failure_cnt += 1
        else:
            downloaded_files_list.append(output_path)
    return downloaded_files_list


def prepare_time_output(times: list[tuple], prefix: str, file: str) -> str:
    """
    Create output directory using:
        OUTPUT_DIR/YEAR/DAY_OF_YEAR/

    Parameters
    ----------
    times : list
        list of tuples consisting of granule names and start times
    prefix : string
        prefix for output path, either custom output -d or short name
    file : string
        granule file name

    Returns
    -------
    write_path
        string path to where granules will be written
    """

    time_match = [_dt for _dt in times if _dt[0] == os.path.splitext(os.path.basename(file))[0]]

    # Found on 11/11/21
    # https://github.com/podaac/data-subscriber/issues/28
    # if we don't find the time match array, try again using the
    # filename AND its suffix (above removes it...)
    if len(time_match) == 0:
        time_match = [_dt for _dt in times if _dt[0] == os.path.basename(file)]
    time_match = time_match[0][1]

    year = time_match.strftime('%Y')
    month = time_match.strftime('%m')
    day = time_match.strftime('%d')
    day_of_year = time_match.strftime('%j')

    time_dir = os.path.join(year, day_of_year)
    pa.check_dir(os.path.join(prefix, time_dir))
    write_path = os.path.join(prefix, time_dir, os.path.basename(file))
    return write_path


def download_cyg_files_between_date(st_date: dt.date, end_date: dt.date, list_sc_num: Optional[Union[list[int], int]] = None,
                                    cyg_data_ver: Optional[str] = None, cyg_data_lvl: str = 'L1', cygnss_l1_path: Optional[str] = None,
                                    checksum_exist_file: bool = False, force_download: bool = False):
    """
    download CYGNSS data between two dates (including start and end date)

    :param st_date: start date
    :type st_date: date
    :param end_date: end date
    :type end_date: date
    :param list_sc_num: list of cygnss spacecraft numbers (1-8), if None will download all SCs
    :type list_sc_num: list or np.array or int
    :param cyg_data_ver: cygnss data version (ex: 'v2.1')
    :type cyg_data_ver: str
    :param cyg_data_lvl: cygnss data level (ex: 'L1')
    :type cyg_data_lvl: str
    :param cygnss_l1_path: path of the cygnss L1 data (default: os.environ.get('CYGNSS_L1_PATH'))
    :type cygnss_l1_path: str or None
    :param checksum_exist_file: check md5 checksum for existing files? this will make it very slow
    :type checksum_exist_file: bool
    :param force_download: download the file even if the version is not included in the path (not recommended)
    :return:
    """

    if cygnss_l1_path is None:
        cygnss_l1_path = os.environ.get('CYGNSS_L1_PATH')
    if list_sc_num is None:
        list_sc_num = np.arange(1, 9)
    elif np.size(list_sc_num) == 1:
        list_sc_num = [int(list_sc_num)]
    search_keywords = [f'cyg{sc_num:02d}' for sc_num in list_sc_num]
    if cygnss_l1_path is None:
        raise ValueError("$CYGNSS_L1_PATH environment variable need to be set, or use cygnss_l1_path input parameter")

    if cyg_data_ver is None:
        folder_list = cygnss_l1_path.split(os.path.sep)
        if not folder_list[-1]:
            folder_list.pop(-1)
        cyg_data_ver = folder_list[-1]
    cyg_data_lvl = cyg_data_lvl.upper()

    # get temporal limit for the search
    start_date_time = st_date.isoformat() + 'T00:00:00Z'
    end_date_time = end_date.isoformat() + 'T23:59:59Z'
    temporal_range = pa.get_temporal_range(start_date_time, end_date_time, dt.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))

    # Get CYGNSS Shortname
    cyg_data_shortname = cyg_ver2shortname(cyg_data_ver, cyg_data_lvl)

    pa.setup_earthdata_login_auth(pa.edl)
    token = pa.get_token(pa.token_url)

    params = [
        ('page_size', page_size),
        ('sort_key', "start_date"),
        ('provider', 'POCLOUD'),
        ('ShortName', cyg_data_shortname),
        ('temporal', temporal_range),
        ('token', token),
    ]

    # If 401 is raised, refresh token and try one more time
    try:
        results = pa.get_search_results(params)
    except HTTPError as e:
        if e.code == 401:
            token = pa.refresh_token(token)
            # Updated: This is not always a dictionary...
            # in fact, here it's always a list of tuples
            for i, p in enumerate(params):
                if p[0] == "token":
                    params[i] = ("token", token)
            results = pa.get_search_results(params)
        else:
            raise e

    checksums = pa.extract_checksums(results)
    file_start_times = pa.parse_start_times(results)

    downloads = []
    for r in results['items']:
        for u in r['umm']['RelatedUrls']:
            if u['Type'] == "GET DATA" and ('Subtype' not in u or u['Subtype'] != "OPENDAP DATA"):
                f = u['URL']
                for extension in ["\\.nc"]:
                    if pa.search_extension(extension, f):
                        for keyword in search_keywords:
                            if keyword in f.lower():
                                downloads.append(f)
                                break
    ver_folder_exist = check_ver_folder(cygnss_l1_path, cyg_data_ver)
    if not force_download and not ver_folder_exist:
        error_str = f"You are trying to download version {cyg_data_ver:s}, but the path doesn't contain the version name"
        raise ValueError(f'{error_str:s}\nuse force_download=True to remove this error')

    success_cnt = failure_cnt = skip_cnt = 0
    downloaded_files_list = []

    for f in downloads:
        try:
            output_path = prepare_time_output(file_start_times, cygnss_l1_path, f)

            # decide if we should actually download this file (e.g. we may already have the latest version)
            if os.path.exists(output_path) and not force_download and (not checksum_exist_file or pa.checksum_does_match(output_path, checksums)):
                logging.info(str(dt.datetime.now()) + " SKIPPED: " + f)
                skip_cnt += 1
                continue

            if len(list(signature(pa.download_file).parameters)) > 3:  # allow for the use of non MiXIL version
                checksum_val = checksums.get(os.path.basename(output_path))
                pa.download_file(f, output_path, checksum_val)
            else:
                pa.download_file(f, output_path)

            logging.info(str(dt.datetime.now()) + " SUCCESS: " + f)
            success_cnt += 1

        except Exception:
            logging.warning(str(dt.datetime.now()) + " FAILURE: " + f, exc_info=True)
            failure_cnt += 1
        else:
            downloaded_files_list.append(output_path)
    return downloaded_files_list


def get_cyg_calibrated_rawif_file(cyg_day_folder: str, sc_num, sampling_rate: float) -> Optional[str]:
    """ Get file name if exist from rawif with specific sampling rate"""
    result = []
    pattern = f"cyg{sc_num:02d}*_rate_{sampling_rate:.3f}*.nc"
    for root, dirs, files in os.walk(cyg_day_folder):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(name)
                break  # finding the first file

    if len(result) == 0:
        cyg_file_name = None
    else:
        cyg_file_name = result[0]
    return cyg_file_name


def get_cyg_file(cyg_day_folder: str, sc_num: int) -> Optional[str]:
    """
    check if the file exist and return the file name, if not exist return None

    :param cyg_day_folder: cygnss day folder
    :type cyg_day_folder: str
    :param sc_num: spacecraft number
    :type sc_num: int
    :return: file name
    :rtype: str
    """
    result = []
    pattern = f"cyg{sc_num:02d}*.nc"
    for root, dirs, files in os.walk(cyg_day_folder):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(name)
                break  # finding the first file

    if len(result) == 0:
        cyg_file_name = None
    else:
        cyg_file_name = result[0]

    return cyg_file_name


def check_ver_folder(cygnss_l1_path: str, cyg_data_ver: str) -> bool:
    """
    check if the version name in the path

    :param cygnss_l1_path: path of the cygnss L1 data
    :type cygnss_l1_path: str
    :param cyg_data_ver: cygnss data version (ex: 'v2.1')
    :type cyg_data_ver: str
    :return:
    """
    path_split = cygnss_l1_path.split(os.sep)
    if cyg_data_ver in path_split:
        out = True
    else:
        warnings.warn(f"You are trying to download version {cyg_data_ver:s}, but the path doesn't contain the version name", RuntimeWarning)
        out = False
    return out


if __name__ == "__main__":

    st_date = dt.date(year=2020, month=7, day=1)
    end_date = dt.date(year=2020, month=12, day=31)
    cyg_data_ver = 'v3.1'
    download_cyg_files(2019, [5, 30, 31, 40])
    download_cyg_files_between_date(st_date, end_date, cyg_data_ver=cyg_data_ver, checksum_exist_file=False)



#!/usr/bin/env python3
"""
 DESCRIPTION
          This tool is part of cygnsslib python package. The package is created by Mixil lab at USC
          See <https://bitbucket.org/usc_mixil/cygnss-library>

          This Tool download SRTM .hgt data

 AUTHOR   Amer Melebari
          Microwave Systems, Sensors and Imaging Lab (MiXiL)
          University of Southern California
 EMAIL    amelebar@usc.edu
 CREATED  2020-08-16
 Updated  2023-05-02


part of the download_file() is from the following codes:
https://git.earthdata.nasa.gov/projects/LPDUR/repos/daac_data_download_python/
https://github.com/DHI-GRAS/earthdata-download/tree/master/earthdata_download

  Copyright 2020 University of Southern California
"""
import logging
import netrc
from time import sleep
from tqdm.auto import tqdm
import numpy as np
import os
import requests
import shutil
import zipfile

CHUNK_SIZE = 512 * 1024  # 512 kB
MAX_CONNECTIONS_TRIES = 10
MIN_FILE_SIZE_BYTES = 10e3
SRTM_USGS_URL = 'https://e4ftl01.cr.usgs.gov/MEASURES/SRTMGL1.003/2000.02.11/'
"""
"""


class EarthdataSession(requests.Session):

    AUTH_DOMAINS = ['nasa.gov', 'usgs.gov']

    def __init__(self, username, password):
        """Create Earthdata Session that preserves headers when redirecting"""
        super(EarthdataSession, self).__init__()  # Python 2 and 3
        self.auth = (username, password)

    def rebuild_auth(self, prepared_request, response):
        """Keep headers upon redirect as long as we are on any of self.AUTH_DOMAINS"""
        headers = prepared_request.headers
        url = prepared_request.url
        if 'Authorization' in headers:
            original_parsed = requests.utils.urlparse(response.request.url)
            redirect_parsed = requests.utils.urlparse(url)
            original_domain = '.'.join(original_parsed.hostname.split('.')[-2:])
            redirect_domain = '.'.join(redirect_parsed.hostname.split('.')[-2:])
            if original_domain != redirect_domain and redirect_domain not in self.AUTH_DOMAINS and original_domain not in self.AUTH_DOMAINS:
                del headers['Authorization']


def _cyg_data_auth_err_handling():
    print('Error in the Earthdata Credentials, try to enter them again.')
    print('Note: you need to re-start the code after this download as the new user/pass need to be re-loaded')
    save_pass_str = input('Do you want to save the new user/pass? (if yes, the old user/pass will be deleted) [Yes, No]')
    save_pass = True if (save_pass_str.lower() in ['y', 'yes']) else False
    podaac_usr, podaac_pass = get_earthdata_cred()
    return podaac_usr, podaac_pass


def _download_file_earthdata(file_url, output_folder, auth):
    """
    download the file with url into folder output_folder

    :param file_url: url of the file
    :type file_url: str
    :param output_folder: saving folder
    :type output_folder: str
    :param auth: username and pass
    :type auth: tuple of str
    :return: downloaded file path
    :rtype: str
    """
    file_name = file_url.split('/')[-1]
    out_file = os.path.join(output_folder, file_name)

    out_file_temp = f'{out_file:s}.incomplete'
    for i_try in np.arange(MAX_CONNECTIONS_TRIES):
        try:
            with EarthdataSession(username=auth[0], password=auth[1]) as session:
                with session.get(file_url, stream=True) as response:
                    if response.status_code == 404:
                        print(f'SRTM download Error: {file_url} is not Found')
                        return None
                    response.raise_for_status()
                    response.raw.decode_content = True
                    with tqdm.wrapattr(open(out_file_temp, "wb"), "write", miniters=1, total=int(response.headers.get('content-length', 0)),
                                       desc=out_file_temp) as target_file:
                        for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                            target_file.write(chunk)

            filesize = os.path.getsize(out_file_temp)
            if filesize < MIN_FILE_SIZE_BYTES:
                raise RuntimeError(f'Size of downloaded file is only {(filesize * 1e-3):.3f} B. Suspecting broken file ({out_file_temp}).')

        except (requests.exceptions.ConnectionError, requests.exceptions.HTTPError) as e:
            if i_try < MAX_CONNECTIONS_TRIES:
                print('Connection Error sleep for 5 sec')
                sleep(5)
            else:
                raise RuntimeError(f'a connection error was raised when trying to access {file_url:s}\n {e}')
        except requests.exceptions.RequestException:
            if i_try < MAX_CONNECTIONS_TRIES:
                podaac_usr, podaac_pass = _cyg_data_auth_err_handling()
                auth = (podaac_usr, podaac_pass)
            else:
                raise RuntimeError('Authentication error, please check your PODAAC user/pass')
        except RuntimeError as e:
            if i_try < MAX_CONNECTIONS_TRIES:
                continue
            else:
                raise RuntimeError(e)

        else:  # if it was successful
            break

    shutil.move(out_file_temp, out_file)

    return out_file


def extract_srtm_zipfile(zip_file, out_path):
    """
    unzip file

    :param zip_file: zip file path
    :type zip_file: str
    :param out_path: extraction out folder
    :type out_path: str
    :return: unzipped file name
    :rtype: str
    """

    with zipfile.ZipFile(zip_file, "r") as zip_ref:
        file_name = zip_ref.namelist()[0]
        zip_ref.extractall(out_path, members=[file_name])

    return os.path.join(out_path, file_name)


def get_srtm_file_name(lat, lon):
    """
    get SRTM file name. the name of the zipped file, for the unzipped file use get_srtm_ght_file_name_frm_lat_lon() or get_srtm_unzipped_file_name()
    :param lat: latitude
    :type lat: int or float
    :param lon: longitude
    :type lon: int or float
    :return:
    """
    file_coord = _get_srtm_file_coord_name(lat, lon)
    file_name = f'{file_coord:s}.SRTMGL1.hgt.zip'
    return file_name


def _get_srtm_file_coord_name(lat, lon):
    """

    :param lat: latitude
    :type lat: int or float
    :param lon: longitude
    :type lon: int or float
    :return:
    """
    lat = int(lat)
    lon = int(lon)
    if lat > 59 or lat < -56:
        raise ValueError('Latitude value should be between -56 to 59. Got {:d}'.format(lat))
    if lon > 179 or lon < -180:
        raise ValueError('Longitude value should be between -180 to 179. Got {:d}'.format(lon))
    lat_direction = 'N' if (lat >= 0) else 'S'
    lon_direction = 'E' if (lon >= 0) else 'W'
    file_coord = '{:s}{:02d}{:s}{:03d}'.format(lat_direction, abs(lat), lon_direction, abs(lon))
    return file_coord


def get_srtm_ght_file_name(lat, lon):
    """
    return the name of the .hgt file from lat, lon value
    :param lat: latitude
    :type lat: int or float
    :param lon: longitude
    :type lon: int or float
    :return:
    """
    file_coord = _get_srtm_file_coord_name(lat, lon)
    file_name = f'{file_coord:s}.hgt'
    return file_name


def get_srtm_unzipped_file_name(zipped_filename):
    """
    return the unzipped SRTM file name from the zipped SRTM file name

    :param zipped_filename: zipped SRTM file name from get_srtm_file_name()
    :type zipped_filename: str
    :return: unzipped SRTM file name
    :rtype: str
    """
    return '{:s}.hgt'.format(zipped_filename.split('.')[0])


def get_earthdata_cred():
    """
    import EarthData username and pass from the system, if not found it ask you to enter them

    :return:
    """
    edl = "urs.earthdata.nasa.gov"
    earthdata_usr = ''
    earthdata_pass = ''
    try:
        earthdata_usr, _, earthdata_pass = netrc.netrc().authenticators(edl)
    except (FileNotFoundError, TypeError):
        # FileNotFound = There's no .netrc file
        # TypeError = The endpoint isn't in the netrc file,
        #  causing the above to try unpacking None
        logging.warning("There's no .netrc file or the The endpoint isn't in the netrc file")

    return earthdata_usr, earthdata_pass


def download_srtm_ght_files(list_lat, list_lon, out_folder=None, earthdata_usr_pass=None, save_pass=True, unzipfile=False, remove_zippedfile=True,
                            silent_run=False):
    """
    download a multiple SRTM file in ght format. it download the files within the lat/long lists

    The files are in this URL: https://e4ftl01.cr.usgs.gov/MEASURES/SRTMGL1.003/2000.02.11/
    the username/password can be obtained from https://urs.earthdata.nasa.gov/home

    :param list_lat: list of latitude
    :type list_lat: list of int or numpy.array or int
    :param list_lon: list of longitude
    :type list_lon: list of int or or numpy.array or int
    :param out_folder: save folder, if None, it will be saved in "$SRTM_PATH/hgt"
    :type out_folder: str
    :param earthdata_usr_pass: earthdata username and password (usr,pass).  https://urs.earthdata.nasa.gov/home
    :type earthdata_usr_pass: tuple of str
    :param save_pass: (active if earthdata_usr_pass is None) save earthdata username and password in the system (select False if you're using a shared
     or public PC)
    :type save_pass: bool
    :param unzipfile: unzip the downloaded file?
    :type unzipfile: bool
    :param remove_zippedfile: delete the unzipped file after unzipping?
    :type remove_zippedfile: bool
    :param silent_run: if true no screen output will be shown except downloading progress and errors
    :type silent_run: bool
    :return: files path
    :rtype: list of str
    """
    try:
        iter(list_lat)
    except TypeError:
        list_lat = [int(list_lat)]
    try:
        iter(list_lon)
    except TypeError:
        list_lon = [int(list_lon)]

    if out_folder is None:
        out_folder = os.path.join(os.environ['SRTM_PATH'], 'hgt')
        os.makedirs(out_folder, exist_ok=True)
    if earthdata_usr_pass is None:
        earthdata_usr_pass = get_earthdata_cred()

    d_file_paths = list()
    for lat in list_lat:
        for lon in list_lon:
            file_name = download_srtm_ght_single_file(lat, lon, out_folder, earthdata_usr_pass, save_pass, unzipfile, remove_zippedfile, silent_run)
            if file_name is not None:  # only add when we download files
                d_file_paths.append(file_name)

    return d_file_paths


def download_srtm_ght_single_file(lat, lon, out_folder=None, earthdata_usr_pass=None, save_pass=True, unzipfile=False, remove_zipedfile=True,
                                  silent_run=False):
    """

    download a single SRTM file in ght format, don't use this function directly, consider using download_srtm_ght_files()
    The files are in this URL: https://e4ftl01.cr.usgs.gov/MEASURES/SRTMGL1.003/2000.02.11/
    the username/password can be obtain from https://urs.earthdata.nasa.gov/home

    :param lat: file latitude
    :type lat: int
    :param lon: file longitude
    :type lon: int
    :param out_folder: save folder, if None, it will be saved in "$SRTM_PATH/hgt"
    :type out_folder: str
    :param earthdata_usr_pass: earthdata username and password (usr,pass).  https://urs.earthdata.nasa.gov/home
    :type earthdata_usr_pass: tuple of string
    :param save_pass: (active if earthdata_usr_pass is None) save earthdata username and password in the system (select False if you're using a shared
     or public PC)
    :type save_pass: bool
    :param unzipfile: unzip the downloaded file?
    :type unzipfile: bool
    :param remove_zipedfile: delete the unzipped file after unzipping? ignored if unzipfile is False
    :type remove_zipedfile: bool
    :param silent_run: if true no screen output will be shown except downloading progress
    :type silent_run: bool
    :return: file path
    :rtype: str
    """
    if out_folder is None:
        out_folder = os.path.join(os.environ['SRTM_PATH'], 'hgt')
        os.makedirs(out_folder, exist_ok=True)

    file_name = get_srtm_file_name(lat, lon)
    file_path = os.path.join(out_folder, file_name)
    unzipped_file_path = os.path.join(out_folder, get_srtm_unzipped_file_name(file_name))
    if os.path.exists(unzipped_file_path):
        if not silent_run:
            print(f'{unzipped_file_path:s} file exist')
        return unzipped_file_path
    if os.path.exists(file_path):
        if not silent_run:
            print(f'{file_path:s} file exist')
        return file_path

    if earthdata_usr_pass is None:
        earthdata_usr_pass = get_earthdata_cred()
    file_url = f'{SRTM_USGS_URL:s}{file_name:s}'
    saved_zip_file = _download_file_earthdata(file_url, out_folder, auth=earthdata_usr_pass)
    if saved_zip_file is None:  # If no file, i.e. over ocean
        return None
    if not unzipfile:
        return saved_zip_file

    extract_srtm_zipfile(saved_zip_file, out_folder)
    unzipped_file = os.path.join(out_folder, get_srtm_unzipped_file_name(file_name))
    if remove_zipedfile:
        os.remove(saved_zip_file)
    return unzipped_file


if __name__ == '__main__':
    files_lat = np.array([21, 22, 20])
    files_lon = [29, 30]
    download_srtm_ght_files(files_lat, files_lon)




from collections import defaultdict
from datetime import datetime
from matplotlib.ticker import AutoMinorLocator, MultipleLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable
from typing import Optional
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd


def axis_formatter_degree(x, pos): return f'{x:-0.0f}'.replace('-', u'\u2212')+ '$^\mathrm{o}$'
try:
    mpl.use('Qt5Agg')
except Exception:
    pass

sel_colormap = 'jet'  # matplotlib default is 'viridis'
default_fig_width = 8  # default figure width
plt_font_size = 20  # font size
plt_title = True  # plot figures titles
plot_db_range = 20  # default dynamic range of dB plots
default_save_type = ['png']  # default save image types
ddm_plt_default_delay_axis = 'x'

SMALL_SIZE = plt_font_size - 2
MEDIUM_SIZE = plt_font_size
BIGGER_SIZE = plt_font_size
plt.rc('font', size=SMALL_SIZE)  # controls default text sizes
plt.rc('axes', titlesize=SMALL_SIZE)  # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

mpl.rcParams['font.size'] = plt_font_size
mpl.rcParams['legend.fontsize'] = 'large'
mpl.rcParams['figure.titlesize'] = 'medium'
mpl.rcParams["axes.unicode_minus"] = True

def lin2db(val):
    if isinstance(val, np.ndarray):
        val[val<=0.0] = np.nan
    else:
        if val <= 0.0:
            val = np.nan
    return 10.0 * np.log10(val)
def axis_formatter_no_frac(x, pos): return f'{x:-0.0f}'.replace('-', u'\u2212')

def format_range_val(start, finish, unit=''):
    return f'[{start}{unit},{finish}{unit}]'.replace('-', u'\u2212')
def pwr2db_threshold(power_linear, dynamic_range_db=None):
    """

    change power from linear two dB with a minimum threshold

    :param power_linear: power in linear scale
    :type power_linear: np.array
    :param dynamic_range_db: dynamic range in dB any value below the maximum by this is set to max - scale
    :type dynamic_range_db: float
    :return: power in dB scale
    :rtype: np.array
    """
    if dynamic_range_db is None:
        dynamic_range_db = plot_db_range

    threshold = np.nanmax(power_linear) * 10 ** (-dynamic_range_db / 10)
    power_db = 10.0 * np.log10(np.where(power_linear < threshold, threshold, power_linear))
    power_db = np.nan_to_num(power_db, nan=np.nanmin(power_db))
    return power_db


def plot_single_ddm(image, title, img_save_name, fig_out_folder, tf_save_fig=True, img_ext=None, fig_width=None,
                    plt_delay_axis=None, cbar_min_max=None, fig_save_types: Optional[list[str]] = None, cbar_title=None,
                    delay_scale=None, dopp_scale=None, plt_db_tf = True):
    """

    plot a single DDM

    :param image: DDM [dim 0: delay, dim 1: Doppler]
    :type image: np.array
    :param title: plot title
    :type title: str
    :param img_save_name: image save name
    :type img_save_name: str
    :param fig_out_folder: image saving folder
    :type fig_out_folder: str
    :param tf_save_fig: save the figure?
    :type tf_save_fig: bool
    :param img_ext: extend of the image
    :type img_ext: tuple
    :param fig_width: figure width, if None default_fig_width is selected
    :type fig_width: int or float
    :param plt_delay_axis: where the delay axis? select 'x' or 'y'
    :type plt_delay_axis: str or None
    :return: figure handle
    :rtype: plt.figure
    """
    plt_bar_title = True
    if delay_scale is None:
        delay_scale = 1.0
    if dopp_scale is None:
        dopp_scale = 1.0
    if cbar_title is None:
        cbar_title = 'Reflectivity [dB]'
    if plt_delay_axis is None:
        plt_delay_axis = ddm_plt_default_delay_axis
    if fig_width is None:
        fig_width = default_fig_width
    if plt_delay_axis.lower() == 'x':
        image = np.transpose(image)

    delay_axis = 1 if (plt_delay_axis.lower() == 'x') else 0
    dopp_axis = 0 if (plt_delay_axis.lower() == 'x') else 1
    num_delay = int(image.shape[delay_axis] * delay_scale / 2)
    num_dopp = int(image.shape[dopp_axis] * dopp_scale / 2)
    if plt_delay_axis.lower() == 'x':
        fig_size = (fig_width, np.round(fig_width * num_dopp / num_delay, decimals=2))
        y_label = 'Doppler bin'
        x_label = 'Delay bin'
        if img_ext is None:  # image ext (left, right, bottom, top)
            img_ext = (-num_delay-0.5*delay_scale,
                       num_delay+0.5*delay_scale,
                       -num_dopp-0.5*dopp_scale,
                       num_dopp+0.5*dopp_scale)
    elif plt_delay_axis.lower() == 'y':
        fig_size = (np.round(fig_width * num_dopp / num_delay, decimals=2), fig_width * 0.9)
        x_label = 'Doppler bin'
        y_label = 'Delay bin'
        if img_ext is None:  # image ext (left, right, bottom, top)
            img_ext = (-num_dopp-0.5,
                       num_dopp+0.5,
                       -num_delay-0.5,
                       num_delay+0.5)
    else:
        raise RuntimeError(f'plt_delay_axis has only two options: x and y, you selected {plt_delay_axis}')

    fig = plt.figure(figsize=fig_size)
    ax = plt.subplot(111)
    im = ax.imshow(image, origin='lower', extent=img_ext, cmap=sel_colormap)
    minor_tick_num = 2
    if cbar_min_max is not None:
        z_min, z_max = cbar_min_max
        if (z_max - z_min) < 15.0:
            z_step = 2
            minor_tick_num = 2
        elif (z_max - z_min) < 30.0:
            z_step = 5
            minor_tick_num = 5
        else:
            z_step = 10
            minor_tick_num = 2
        cbar_ticks = np.arange(z_min, z_max + z_step, z_step)
        im.set_clim(z_min, z_max)
    elif plt_db_tf:
        z_min = np.nanmin(image)
        z_max = np.nanmax(image)
        if (z_max - z_min) < 15.0:
            z_step = 2
            minor_tick_num = 2
        elif (z_max - z_min) < 30.0:
            z_step = 5
            minor_tick_num = 5
        else:
            z_step = 10
            minor_tick_num = 2
        _z_step = np.minimum(5, z_step)
        vmin = np.floor(z_min / _z_step) * _z_step
        vmax = np.ceil(z_max / _z_step) * _z_step
        try:
            cbar_ticks = np.arange(vmin, vmax + z_step, z_step)
        except ValueError:
            print(f'Error in setting cbar, {vmin}, {vmax}')
            cbar_ticks = None

        im.set_clim(vmin, vmax)
    else:
        minor_tick_num = 2
        cbar_ticks = None
    if title and plt_title:  # this in case we want to remove titles for papers
        plt.title(title, fontsize=plt_font_size-2)
    plt.xlabel(x_label, fontsize=plt_font_size)
    plt.ylabel(y_label, fontsize=plt_font_size)
    ax.tick_params(axis='both', which='major', labelsize=plt_font_size)
    if num_delay % 2 != 0:  # odd number
        num_delay += -1
    if num_dopp % 2 != 0:  # odd number
        num_dopp += -1
    delay_step = 2 if num_delay < 10 else 4
    delay_ticks = np.arange(-num_delay, stop=num_delay + delay_step, step=delay_step).astype(int)
    dopp_step = 2 if num_dopp < 10 else 4
    dopp_ticks = np.arange(-num_dopp, stop=num_dopp + dopp_step, step=dopp_step).astype(int)
    if plt_delay_axis.lower() == 'x':
        ax.set_xticks(delay_ticks)
        ax.set_yticks(dopp_ticks)
    else:
        ax.set_yticks(delay_ticks)
        ax.set_xticks(dopp_ticks)
    ax.xaxis.set_minor_locator(AutoMinorLocator(2))
    ax.yaxis.set_minor_locator(AutoMinorLocator(2))

    # create an axes on the right side of ax. The width of cax will be 5%
    # of ax and the padding between cax and ax will be fixed at 0.05 inch.
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.1)
    cbar = plt.colorbar(im, cax=cax)
    # cbar = plt.colorbar(fraction=c_fraction, pad=0.02)
    cbar.ax.tick_params(labelsize=plt_font_size)
    cbar.ax.yaxis.set_major_formatter(axis_formatter_no_frac)
    # Turn on minor ticks
    if cbar_ticks is not None:
        cbar.set_ticks(cbar_ticks)
    cbar.minorticks_on()
    cbar.ax.yaxis.set_minor_locator(AutoMinorLocator(minor_tick_num))  # For vertical colorbar

    if plt_bar_title:
        cbar.ax.set_ylabel(cbar_title, rotation=270, fontsize=plt_font_size, labelpad=plt_font_size)
    plt.tight_layout()
    plt.tight_layout()
    plt.tight_layout()
    save_figure(fig, fig_out_folder, img_save_name, tf_save_fig, fig_save_types)
    return fig


# Function to plot data for varying sp_inc_angle with fixed other parameters
def plot_varying_sp_inc_angle(grouped_data, folder_path, img_tag, fig_save_types):
    for smap_sm_group, sc_alt_groups in grouped_data.items():
        for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
            list_az_angles = []
            for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                for sc_az_angle_group, items in sc_az_angle_groups.items():
                    if sc_az_angle_group not in list_az_angles:
                        list_az_angles.append(sc_az_angle_group)
            list_az_angles.sort()
            for sc_az_angle_group in list_az_angles:
                fig = plt.figure()
                ax = plt.subplot(111)
                i_plt = 0
                for sp_inc_angle_group, sc_az_angle_items in sp_inc_angle_groups.items():
                    if sc_az_angle_group in sc_az_angle_items:
                        items = sc_az_angle_items[sc_az_angle_group]

                        timestamps = [datetime.fromisoformat(item['ddm_timestamp_utc_str'][:-3]) for item in items]
                        reflectivity_peaks = [item['reflectivity_peak'] for item in items]

                        if len(reflectivity_peaks) < 2:
                            continue
                        # Sort data by timestamps
                        sorted_data = sorted(zip(timestamps, reflectivity_peaks))
                        sorted_timestamps, sorted_reflectivity_peaks = zip(*sorted_data)
                        peak_refl = lin2db(np.array(sorted_reflectivity_peaks))
                        ax.plot(sorted_timestamps, peak_refl, marker='o', linestyle='', label=format_range_val(sp_inc_angle_group, sp_inc_angle_group+5, '$^\mathrm{o}$'))
                        i_plt += 1

                if i_plt == 0:
                    plt.close(fig)
                    continue
                ax.set_xlabel('Timestamp (UTC)', fontsize=plt_font_size)
                ax.set_ylabel('Peak reflectivity [dB]', fontsize=plt_font_size)
                ax.grid()
                # Determine the tick range for y-axis
                y_min, y_max = plt.ylim()
                # Define ticks with a step of y-axis
                if (y_max - y_min) < 15.0:
                    y_step = 2
                    num_minor_ticks = 2
                elif (y_max - y_min) < 30.0:
                    y_step = 5
                    num_minor_ticks = 5
                else:
                    y_step = 10
                    num_minor_ticks = 2

                y_ticks = np.arange(np.floor(y_min/y_step) * y_step, np.ceil(y_max/y_step) * y_step + y_step, y_step)
                ax.set_yticks(y_ticks)
                ax.yaxis.set_minor_locator(AutoMinorLocator(num_minor_ticks))

                # plt.title(f'Reflectivity Peak over Time for smap_sm: {smap_sm_group}, sc_alt: {sc_alt_group}, sc_az_angle: {sc_az_angle_group}')
                legend_vbox = 1.5

                plt.legend(title='$\\theta_\mathrm{i}$', loc="upper center", bbox_to_anchor=(.5, legend_vbox), ncol=3,
                           fontsize=plt_font_size - (4 if plt_font_size > 16 else 0), handletextpad=0.3, borderpad=0.3, labelspacing=0.3, columnspacing=1.0, markerscale=1.5)
                plt.xticks(rotation=45)
                sc_az_angle_group_str = str(sc_az_angle_group) if sc_az_angle_group > 0.0 else 'n' + str(abs(sc_az_angle_group))
                img_name = img_tag + f'_peak_refl_vary_inc_angle_sm_{int(smap_sm_group * 100):d}_alt_{sc_alt_group}_az_{sc_az_angle_group_str}'
                save_figure(fig, folder_path, img_name, fig_save_types=fig_save_types)
                plt.close(fig)


# Function to plot data for varying sc_az_angle with fixed other parameters
def plot_varying_sc_az_angle(grouped_data, folder_path, img_tag, fig_save_types):
    for smap_sm_group, sc_alt_groups in grouped_data.items():
        for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
            for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                fig = plt.figure()
                ax = plt.subplot(111)
                i_plt = 0
                for sc_az_angle_group, items in sc_az_angle_groups.items():
                    timestamps = [datetime.fromisoformat(item['ddm_timestamp_utc_str'][:-3]) for item in items]
                    reflectivity_peaks = [item['reflectivity_peak'] for item in items]

                    # Sort data by timestamps
                    sorted_data = sorted(zip(timestamps, reflectivity_peaks))
                    sorted_timestamps, sorted_reflectivity_peaks = zip(*sorted_data)
                    peak_refl = lin2db(np.array(sorted_reflectivity_peaks))
                    ax.plot(sorted_timestamps, peak_refl, marker='o', linestyle='', label=format_range_val(sp_inc_angle_group, sc_az_angle_group+5, '$^\mathrm{o}$'))
                    i_plt += 1
                if i_plt == 0:
                    plt.close(fig)
                    continue

                ax.set_xlabel('Timestamp (UTC)')
                ax.set_ylabel('Peak reflectivity [dB]')
                ax.grid()

                # Determine the tick range for y-axis
                y_min, y_max = plt.ylim()
                # Define ticks with a step of y-axis
                if (y_max - y_min) < 15.0:
                    y_step = 2
                    num_minor_ticks = 2
                elif (y_max - y_min) < 30.0:
                    y_step = 5
                    num_minor_ticks = 5
                else:
                    y_step = 10
                    num_minor_ticks = 2

                y_ticks = np.arange(np.floor(y_min/y_step) * y_step, np.ceil(y_max/y_step) * y_step + y_step, y_step)
                ax.set_yticks(y_ticks)
                ax.yaxis.set_minor_locator(AutoMinorLocator(num_minor_ticks))

                # plt.title(f'Reflectivity Peak over Time for smap_sm: {smap_sm_group}, sc_alt: {sc_alt_group}, sp_inc_angle: {sp_inc_angle_group}')
                legend_vbox = 1.5
                plt.legend(title='$\\theta_\mathrm{az}$', loc="upper center", bbox_to_anchor=(.5, legend_vbox), ncol=3,
                           fontsize=plt_font_size - (4 if plt_font_size > 16 else 0), handletextpad=0.3, borderpad=0.3, labelspacing=0.3, columnspacing=1.0, markerscale=1.5)

                plt.xticks(rotation=45)
                plt.tight_layout()
                img_name = img_tag + f'_peak_refl_vary_az_angle_sm_{int(smap_sm_group * 100):d}_alt_{sc_alt_group}_az_{sp_inc_angle_group}'
                save_figure(fig, folder_path, img_name, fig_save_types=fig_save_types)
                plt.close(fig)


# Function to plot data for varying sc_az_angle with fixed other parameters
def plot_reflectivity_x_az_lg_inc_angle(grouped_data, folder_path, img_tag, fig_save_types, db_lim=None, separate_legend=True):
    # clr_list = mpl.rcParams['axes.prop_cycle']
    # clr_list = ['#9e0142', '#d53e4f', '#f46d43', '#fdae61', '#fee08b', '#e6f598', '#abdda4', '#66c2a5', '#3288bd', '#5e4fa2']
    clr_list = ["#E5E5E5","#D2D2D2","#BFBFBF","#ACACAC","#999999","#868686","#737373","#5F5F5F","#4C4C4C","#393939","#262626","#131313","#000000"]
    symb_list = ['o', 's', '^']
    save_lim = 3  # save image if number of incidence angles >= savelim
    for smap_sm_group, sc_alt_groups in grouped_data.items():
        if smap_sm_group < 0: # skip soil moisture below 0.
            continue
        for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
            # Sort sp_inc_angle groups before plotting
            sorted_sp_inc_angle_groups = sorted(sp_inc_angle_groups.items(), key=lambda x: x[0])
            num_plt = len([inc_ang for inc_ang, _ in sorted_sp_inc_angle_groups if inc_ang <= 60.0])
            if num_plt < save_lim:
                continue
            n_rows = int(np.ceil(num_plt/3))
            fig_height = _get_fig_height_from_leg_row(n_rows) if not separate_legend else 4.2
            fig = plt.figure(figsize=(6,fig_height))
            ax = plt.subplot(111)
            lines = []
            i_plt = 0
            for sp_inc_angle_group, sc_az_angle_groups in sorted_sp_inc_angle_groups:
                if sp_inc_angle_group > 60.0:  # only plot incidence angles below 60
                    continue
                clr_idx = int(sp_inc_angle_group / 5)
                sc_az_angles = []
                reflectivity_peaks = []

                for sc_az_angle_group, items in sc_az_angle_groups.items():
                    sc_az_angles.extend([item['sc_az_angle'] for item in items])
                    reflectivity_peaks.extend([item['reflectivity_peak'] for item in items])

                # Sort data by sc_az_angles
                sorted_data = sorted(zip(sc_az_angles, reflectivity_peaks))
                sorted_sc_az_angles, sorted_reflectivity_peaks = zip(*sorted_data)
                peak_refl = lin2db(np.array(sorted_reflectivity_peaks))
                l = ax.plot(sorted_sc_az_angles, peak_refl, marker=symb_list[clr_idx % len(symb_list)], linestyle='', color=clr_list[clr_idx % len(clr_list)],
                         label=format_range_val(sp_inc_angle_group, sp_inc_angle_group+5, '$^\mathrm{o}$'))
                lines +=l
                i_plt += 1
            ax.set_xlim([-180, 180])
            xticks = np.arange(-180, 180+90, 90)
            ax.set_xticks(xticks)
            ax.xaxis.set_minor_locator(AutoMinorLocator(2))

            if db_lim is not None:
                ax.set_ylim(db_lim[0], db_lim[1])
            # Determine the tick range for y-axis
            y_min, y_max = plt.ylim()
            # Define ticks with a step of y-axis
            if (y_max - y_min) < 15.0:
                y_step = 2
                num_minor_ticks = 2
            elif (y_max - y_min) < 30.0:
                y_step = 5
                num_minor_ticks = 5
            else:
                y_step = 10
                num_minor_ticks = 2

            y_ticks = np.arange(np.floor(y_min/y_step) * y_step, np.ceil(y_max/y_step) * y_step + y_step, y_step)
            ax.set_yticks(y_ticks)
            ax.yaxis.set_minor_locator(AutoMinorLocator(num_minor_ticks))

            plt.xlabel('Receiver azimuth angle')
            plt.ylabel('Peak reflectivity [dB]')
            ax.xaxis.set_major_formatter(axis_formatter_degree)
            plt.grid()
            if not separate_legend:
                ax.set_position([0.18, 0.14, 0.75, 0.6])  # [left, bottom, width, height] in figure fraction
                legend_vbox = _get_fig_legend_bbox(n_rows)
                plt.legend(title='$\\theta_\mathrm{i}$', loc="upper center", bbox_to_anchor=(.5, legend_vbox), ncol=3,
                           fontsize=plt_font_size - (4 if plt_font_size > 16 else 0), handletextpad=0.0, borderpad=0.3,
                           labelspacing=0.3, columnspacing=0.3, markerscale=1.0)
                fig.set_figheight(_get_fig_height_from_leg_row(n_rows))  # Adjust the total figure height
                plt.tight_layout(rect=[0, 0, 1, 0.95])  # Adjust rect to leave space for the legend
            else:
                plt.tight_layout()
            img_name = img_tag + f'_peak_refl_vs_az_angle_varying_inc_angle_sm_{int(smap_sm_group * 100):d}_alt_{sc_alt_group}'
            if separate_legend:
                img_name += '_plot'
            save_figure(fig, folder_path, img_name, fig_save_types=fig_save_types)

            if separate_legend:
                # Create a new figure for the legend
                leg_h = 1 + 0.3 * n_rows
                fig_legend = plt.figure(figsize=(5, leg_h))  # Adjust size as needed
                fig_legend.legend(handles=lines, loc='center', title='$\\theta_\mathrm{i}$', ncol=3,
                                  fontsize=plt_font_size - (4 if plt_font_size > 16 else 0), handletextpad=0.0, borderpad=0.3, labelspacing=0.3,
                                  columnspacing=0.3, markerscale=1.0)
                # Turn off axes for the legend-only figure
                plt.axis('off')
                plt.tight_layout()
                img_name = img_name[:-4] + 'legend'
                save_figure(fig_legend, folder_path, img_name, fig_save_types=fig_save_types)
                plt.close(fig_legend)

            plt.close(fig)

def _get_fig_height_from_leg_row(n_rows):
    fig_height = 4.2 + 0.55 * n_rows
    return fig_height
def _get_fig_legend_bbox(n_rows):
    return 1.25 + 0.07 * n_rows
# Function to plot data for varying sc_az_angle with fixed other parameters
def plot_reflectivity_x_sm_lg_inc_angle_all_az(grouped_data, folder_path, img_tag, fig_save_types, db_lim=None, separate_legend=True):
    plt_linear_fit = True
    linter_fit_refl_in_db = True
    save_lim = 3  # save image if number of incidence angles >= savelim
    # clr_list = mpl.rcParams['axes.prop_cycle'].by_key()['color']
    clr_list = ["#E5E5E5","#D2D2D2","#BFBFBF","#ACACAC","#999999","#868686","#737373","#5F5F5F","#4C4C4C","#393939","#262626","#131313","#000000"]
    symb_list = ['o', 's', '^']

    list_alt_vals = []
    list_az_angles = []
    linear_fit_csv_fn = img_tag + f'_linear_fit_sm_refl.csv'
    linear_fit_csv_fp = os.path.join(folder_path, linear_fit_csv_fn)
    for smap_sm_group, sc_alt_groups in grouped_data.items():
        for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
            if sc_alt_group not in list_alt_vals:
                list_alt_vals.append(sc_alt_group)
            for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                if sp_inc_angle_group > 60.0:  # only plot incidence angles below 60
                    continue
                for az_angle_group in sc_az_angle_groups:
                    if az_angle_group not in list_az_angles:
                        list_az_angles.append(az_angle_group)
    linear_fit_data = {'altitude': [],
                       'slope_a1': [],
                       'bias_a0': [],
                       'num_samples': []}
    for alt_val in list_alt_vals:
        smap_sm_values = defaultdict(list)
        sp_inc_angle_values = defaultdict(list)
        reflectivity_peaks = defaultdict(list)

        for az_angle in list_az_angles:
            for smap_sm_group, sc_alt_groups in grouped_data.items():
                if alt_val in sc_alt_groups:
                    sp_inc_angle_groups = sc_alt_groups[alt_val]
                    for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                        if sp_inc_angle_group > 60.0:  # only plot incidence angles below 60
                            continue
                        for sc_az_angle_group, items in sc_az_angle_groups.items():
                            # Extract smap_sm, sp_inc_angle, and reflectivity_peak
                            smap_sm_values[sp_inc_angle_group] += [item['smap_sm'] for item in items if item['smap_sm'] > 0.0]
                            sp_inc_angle_values[sp_inc_angle_group] += [item['sp_inc_angle'] for item in items if item['smap_sm'] > 0.0]
                            reflectivity_peaks[sp_inc_angle_group] += [item['reflectivity_peak'] for item in items if item['smap_sm'] > 0.0]

        num_plt = len(reflectivity_peaks)
        n_rows = int(np.ceil(num_plt/3))
        fig_height = _get_fig_height_from_leg_row(n_rows) if not separate_legend else 4.2
        fig = plt.figure(figsize=(6,fig_height))
        ax = plt.subplot(111)
        i_plt = 0
        lines = []
        if num_plt < save_lim:
            continue

        all_refl = np.concatenate([np.array(_refl) for _refl in reflectivity_peaks.values()])
        all_sm = np.concatenate([np.array(_sm) for _sm in smap_sm_values.values()])
        _sm_lin_fit = np.arange(0, 0.4+0.01,0.01)

        if linter_fit_refl_in_db:
            coef_ = np.polynomial.polynomial.polyfit(all_sm, lin2db(all_refl), 1)
            _refl_lin_fit = coef_[1] * _sm_lin_fit + coef_[0]
        else:
            coef_ = np.polynomial.polynomial.polyfit(all_sm, all_refl, 1)
            _refl_lin_fit = lin2db(coef_[1] * _sm_lin_fit + coef_[0])
        linear_fit_data['altitude'].append(alt_val)
        linear_fit_data['slope_a1'].append(coef_[1])
        linear_fit_data['bias_a0'].append(coef_[0])
        linear_fit_data['num_samples'].append(len(all_refl))
        uniq_inc_angle = sorted(smap_sm_values.keys())
        for sp_inc_angle_group in uniq_inc_angle:
            peak_refl = lin2db(np.array(reflectivity_peaks[sp_inc_angle_group]))
            _sm_vals = smap_sm_values[sp_inc_angle_group]
            clr_idx = int(sp_inc_angle_group/5)
            l = ax.plot(_sm_vals, peak_refl, marker=symb_list[clr_idx % len(symb_list)], linestyle='', color=clr_list[clr_idx % len(clr_list)],
                     label=format_range_val(sp_inc_angle_group, sp_inc_angle_group+5, '$^\mathrm{o}$'))
            i_plt += 1
            lines += l
        if plt_linear_fit:
            l0 = ax.plot(_sm_lin_fit, _refl_lin_fit, linestyle='--', color='b', alpha=0.8, label='Linear fit')
            lines += l0
        plt.xlabel('Soil moisture $[\mathrm{m}^3 \mathrm{m}^{-3}]$')
        plt.xlim([0.0, 0.4])
        x_ticks = np.arange(0, 0.6, 0.1)
        ax.set_xticks(x_ticks)
        ax.xaxis.set_minor_locator(AutoMinorLocator(2))

        if db_lim is not None:
            ax.set_ylim(db_lim[0], db_lim[1])
        # Determine the tick range for y-axis
        y_min, y_max = plt.ylim()
        # Define ticks with a step of y-axis
        if (y_max - y_min) < 15.0:
            y_step = 2
            num_minor_ticks = 2
        elif (y_max - y_min) < 30.0:
            y_step = 5
            num_minor_ticks = 5
        else:
            y_step = 10
            num_minor_ticks = 2
        y_ticks = np.arange(np.floor(y_min/y_step) * y_step, np.ceil(y_max/y_step) * y_step + y_step, y_step)
        ax.set_yticks(y_ticks)
        ax.yaxis.set_minor_locator(AutoMinorLocator(num_minor_ticks))
        plt.ylabel('Peak reflectivity [dB]')
        plt.grid()
        if not separate_legend:
            ax.set_position([0.18, 0.14, 0.75, 0.6])  # [left, bottom, width, height] in figure fraction
            legend_vbox = _get_fig_legend_bbox(n_rows)
            plt.legend(title='$\\theta_\mathrm{i}$', loc="upper center", bbox_to_anchor=(.5, legend_vbox), ncol=3,
                       fontsize=plt_font_size - (4 if plt_font_size > 16 else 0), handletextpad=0.0, borderpad=0.3, labelspacing=0.3, columnspacing=0.3, markerscale=1.0)
            fig.set_figheight(_get_fig_height_from_leg_row(n_rows))  # Adjust the total figure height
            plt.tight_layout(rect=[0, 0, 1, 0.95])  # Adjust rect to leave space for the legend
        else:
            plt.tight_layout()
        img_name = img_tag + f'_peak_refl_vs_sm_varying_inc_angle_all_az_alt_{alt_val}'
        if separate_legend:
            img_name += '_plot'
        save_figure(fig, folder_path, img_name, fig_save_types=fig_save_types)

        if separate_legend:
            # Create a new figure for the legend
            leg_h = 1 + 0.3 * n_rows
            fig_legend = plt.figure(figsize=(5, leg_h))  # Adjust size as needed
            fig_legend.legend(handles=lines, loc='center', title='$\\theta_\mathrm{i}$', ncol=3,
                              fontsize=plt_font_size - (4 if plt_font_size > 16 else 0), handletextpad=0.0, borderpad=0.3, labelspacing=0.3,
                              columnspacing=0.3, markerscale=1.0)
            # Turn off axes for the legend-only figure
            plt.axis('off')
            plt.tight_layout()
            img_name = img_name[:-4] + 'legend'
            save_figure(fig_legend, folder_path, img_name, fig_save_types=fig_save_types)
            plt.close(fig_legend)

        plt.close(fig)
    pd.DataFrame(linear_fit_data).to_csv(linear_fit_csv_fp, index=False)


# Function to plot data for varying sc_az_angle with fixed other parameters
def plot_reflectivity_x_az_lg_sm(grouped_data, folder_path, img_tag, fig_save_types, db_lim=None):
    # clr_list = mpl.rcParams['axes.prop_cycle']
    # clr_list = ['#9e0142', '#d53e4f', '#f46d43', '#fdae61', '#fee08b', '#e6f598', '#abdda4', '#66c2a5', '#3288bd', '#5e4fa2']
    clr_list = ["#E5E5E5","#D2D2D2","#BFBFBF","#ACACAC","#999999","#868686","#737373","#5F5F5F","#4C4C4C","#393939","#262626","#131313","#000000"]
    symb_list = ['o', 's', '^']
    save_lim = 3  # save image if number of incidence angles >= savelim
    list_alt_vals = []
    list_inc_angles = []
    for smap_sm_group, sc_alt_groups in grouped_data.items():
        for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
            if sc_alt_group not in list_alt_vals:
                list_alt_vals.append(sc_alt_group)
            for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                if sp_inc_angle_group not in list_inc_angles:
                    list_inc_angles.append(sp_inc_angle_group)

    sorted_sm_data = sorted(grouped_data.items(), key=lambda x: x[0])
    for alt_val in list_alt_vals:
        for inc_angle in list_inc_angles:
            num_plt = len([sm_val for sm_val, _ in sorted_sm_data if sm_val > 0.0])
            if num_plt <= save_lim:
                continue
            n_rows = int(np.ceil(num_plt/3))
            fig_height = _get_fig_height_from_leg_row(n_rows)
            fig = plt.figure(figsize=(6,fig_height))
            ax = plt.subplot(111)
            i_plt = 0
            plotted_sm_val = []
            lines = []
            for smap_sm_group, sc_alt_groups in sorted_sm_data:
                if smap_sm_group < 0: # skip soil moisture below 0.
                    continue
                clr_idx = int(smap_sm_group / 0.05)

                if alt_val in sc_alt_groups:
                    for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
                        if inc_angle in sp_inc_angle_groups:
                            sc_az_angle_groups =  sp_inc_angle_groups[inc_angle]
                            sc_az_angles = []
                            reflectivity_peaks = []

                            for sc_az_angle_group, items in sc_az_angle_groups.items():
                                sc_az_angles.extend([item['sc_az_angle'] for item in items])
                                reflectivity_peaks.extend([item['reflectivity_peak'] for item in items])

                            # Sort data by sc_az_angles
                            sorted_data = sorted(zip(sc_az_angles, reflectivity_peaks))
                            sorted_sc_az_angles, sorted_reflectivity_peaks = zip(*sorted_data)
                            peak_refl = lin2db(np.array(sorted_reflectivity_peaks))
                            sm_rounded = int(np.round(smap_sm_group * 100)) / 100.0
                            upp_lim_grp_sm = int(np.round((sm_rounded+0.05) * 100)) / 100.0
                            l1 = plt.plot(sorted_sc_az_angles, peak_refl, marker=symb_list[clr_idx % len(symb_list)], linestyle='', color=clr_list[clr_idx % len(clr_list)],
                                     label=format_range_val(sm_rounded, upp_lim_grp_sm, ''))
                            if sm_rounded not in plotted_sm_val:
                                plotted_sm_val.append(sm_rounded)
                                lines += l1
                                i_plt += 1
            if i_plt <= save_lim:
                plt.close(fig)
                continue
            ax.set_xlim([-180, 180])
            xticks = np.arange(-180, 180+90, 90)
            ax.xaxis.set_minor_locator(AutoMinorLocator(2))

            if db_lim is not None:
                ax.set_ylim(db_lim[0], db_lim[1])
            # Determine the tick range for y-axis
            y_min, y_max = plt.ylim()

            # Define ticks with a step of y-axis
            if (y_max - y_min) < 15.0:
                y_step = 2
                num_minor_ticks = 2
            elif (y_max - y_min) < 30.0:
                y_step = 5
                num_minor_ticks = 5
            else:
                y_step = 10
                num_minor_ticks = 2

            y_ticks = np.arange(np.floor(y_min/y_step) * y_step, np.ceil(y_max/y_step) * y_step + y_step, y_step)
            ax.set_yticks(y_ticks)
            ax.yaxis.set_minor_locator(AutoMinorLocator(num_minor_ticks))

            ax.set_xticks(xticks)
            plt.xlabel('Receiver azimuth angle')
            plt.ylabel('Peak reflectivity [dB]')
            plt.grid()
            ax.xaxis.set_major_formatter(axis_formatter_degree)
            # plt.gca().xaxis.set_major_formatter(StrMethodFormatter("{x:}°"))

            # plt.title(f'Reflectivity Peak over Time for smap_sm: {smap_sm_group}, sc_alt: {sc_alt_group}, sp_inc_angle: {sp_inc_angle_group}')
            ax.set_position([0.18, 0.14, 0.75, 0.6])  # [left, bottom, width, height] in figure fraction
            legend_vbox = _get_fig_legend_bbox(n_rows)
            labels = [l.get_label() for l in lines]
            plt.legend(lines, labels, title='Soil moisture [$\mathrm{m}^3\mathrm{m}^{-3}$]', loc="upper center", bbox_to_anchor=(.5, legend_vbox), ncol=3,
                       fontsize=plt_font_size - (4 if plt_font_size > 16 else 0), handletextpad=0.0, borderpad=0.3, labelspacing=0.3,
                       columnspacing=0.3, markerscale=1.0)
            fig.set_figheight(_get_fig_height_from_leg_row(n_rows))  # Adjust the total figure height
            plt.tight_layout(rect=[0, 0, 1, 0.95])  # Adjust rect to leave space for the legend

            img_name = img_tag + f'_peak_refl_vs_az_angle_varying_sm_inc_angle_{int(inc_angle):d}_alt_{alt_val}'
            save_figure(fig, folder_path, img_name, fig_save_types=fig_save_types)
            plt.close(fig)


def plot_reflectivity_vs_inc_angle_sm(grouped_data, folder_path, img_tag, fig_save_types):
    list_alt_vals = []
    list_az_angles = []
    for smap_sm_group, sc_alt_groups in grouped_data.items():
        for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
            if sc_alt_group not in list_alt_vals:
                list_alt_vals.append(sc_alt_group)
            for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                for az_angle_group in sc_az_angle_groups:
                    if az_angle_group not in list_az_angles:
                        list_az_angles.append(az_angle_group)

    for alt_val in list_alt_vals:
        for az_angle in list_az_angles:
            smap_sm_values = []
            sp_inc_angle_values = []
            reflectivity_peaks = []
            for smap_sm_group, sc_alt_groups in grouped_data.items():
                for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
                    for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                        for sc_az_angle_group, items in sc_az_angle_groups.items():
                            # Extract smap_sm, sp_inc_angle, and reflectivity_peak
                            smap_sm_values += [item['smap_sm'] for item in items if item['smap_sm'] > 0.0]
                            sp_inc_angle_values += [item['sp_inc_angle'] for item in items if item['smap_sm'] > 0.0]
                            reflectivity_peaks += [item['reflectivity_peak'] for item in items if item['smap_sm'] > 0.0]


            fig = plt.figure(figsize=(6, 4))
            ax = plt.subplot(111)
            # Create scatter plot
            reflectivity_peaks_db = lin2db(np.array(reflectivity_peaks))
            z_min = np.nanmin(reflectivity_peaks_db)
            z_max = np.nanmax(reflectivity_peaks_db)
            if (z_max - z_min) < 15.0:
                z_step = 2
                minor_tick_num = 2
            elif (z_max - z_min) < 30.0:
                z_step = 5
                minor_tick_num = 5
            else:
                z_step = 10
                minor_tick_num = 2

            _z_step = np.minimum(5, z_step)
            vmin = np.floor(z_min / _z_step) * _z_step
            vmax = np.ceil(z_max / _z_step) * _z_step
            scatter = plt.scatter(smap_sm_values, sp_inc_angle_values, c=reflectivity_peaks_db, cmap=sel_colormap,
                                  marker='o', vmin=vmin, vmax=vmax)
            cbar = plt.colorbar(scatter, label='Peak reflectivity [dB]')

            cbar_ticks = np.arange(vmin, vmax + z_step, z_step)
            cbar.set_ticks(cbar_ticks)
            cbar.ax.minorticks_on()
            cbar.ax.yaxis.set_minor_locator(AutoMinorLocator(minor_tick_num))  # For vertical colorbars

            ax.yaxis.set_major_formatter(axis_formatter_degree)
            plt.xlabel('Soil moisture $[\mathrm{m}^3 \mathrm{m}^{-3}]$')
            plt.xlim([0.0, 0.4])
            x_ticks = np.arange(0, 0.6, 0.1)
            ax.set_xticks(x_ticks)
            ax.xaxis.set_minor_locator(AutoMinorLocator(2))

            # Determine the tick range for y-axis

            y_ticks = np.arange(0, 70, 10)
            ax.set_yticks(y_ticks)
            ax.set_ylim([0, 65])
            ax.yaxis.set_minor_locator(AutoMinorLocator(2))

            plt.ylabel('$\\theta_\mathrm{i}$')
            plt.tight_layout()
            plt.tight_layout()
            plt.tight_layout()
            plt.tight_layout()
            plt.tight_layout()
            az_ang_str = 'n' if az_angle < 0 else ''
            az_ang_str += str(abs(int(az_angle)))
            img_name = img_tag + f'_peak_refl_vs_sm_inc_angle_alt_{alt_val}_az_{az_ang_str}'
            save_figure(fig, folder_path, img_name, fig_save_types=fig_save_types)
            plt.close(fig)



def plot_reflectivity_vs_inc_angle_sm_all_az(grouped_data, folder_path, img_tag, fig_save_types):
    list_alt_vals = []
    list_az_angles = []
    for smap_sm_group, sc_alt_groups in grouped_data.items():
        for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
            if sc_alt_group not in list_alt_vals:
                list_alt_vals.append(sc_alt_group)
            for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                for az_angle_group in sc_az_angle_groups:
                    if az_angle_group not in list_az_angles:
                        list_az_angles.append(az_angle_group)

    for alt_val in list_alt_vals:
        smap_sm_values = []
        sp_inc_angle_values = []
        reflectivity_peaks = []

        for az_angle in list_az_angles:
            for smap_sm_group, sc_alt_groups in grouped_data.items():
                for sc_alt_group, sp_inc_angle_groups in sc_alt_groups.items():
                    for sp_inc_angle_group, sc_az_angle_groups in sp_inc_angle_groups.items():
                        for sc_az_angle_group, items in sc_az_angle_groups.items():
                            # Extract smap_sm, sp_inc_angle, and reflectivity_peak
                            smap_sm_values += [item['smap_sm'] for item in items if item['smap_sm'] > 0.0]
                            sp_inc_angle_values += [item['sp_inc_angle'] for item in items if item['smap_sm'] > 0.0]
                            reflectivity_peaks += [item['reflectivity_peak'] for item in items if item['smap_sm'] > 0.0]


        fig = plt.figure(figsize=(6, 4))
        ax = plt.subplot(111)
        # Create scatter plot
        reflectivity_peaks_db = lin2db(np.array(reflectivity_peaks))
        max_dynamic_range = 20
        z_min = np.nanmin(reflectivity_peaks_db)
        z_max = np.nanmax(reflectivity_peaks_db)
        if (z_max - z_min) < 15.0:
            z_step = 2
            minor_tick_num = 2
        elif (z_max - z_min) < 30.0:
            z_step = 5
            minor_tick_num = 5
        else:
            z_step = 5
            minor_tick_num = 2
        _z_step = np.minimum(5, z_step)
        vmin = np.floor(z_min / _z_step) * _z_step
        vmax = np.ceil(z_max / _z_step) * _z_step

        if (vmax - vmin) > max_dynamic_range:  # restrict dynamic range
            z_median = np.nanmedian(reflectivity_peaks_db)
            z_mid = np.floor(z_median / _z_step) * _z_step
            vmin = z_mid - max_dynamic_range/2
            vmax = z_mid + max_dynamic_range/2

        scatter = plt.scatter(smap_sm_values, sp_inc_angle_values, c=reflectivity_peaks_db, cmap=sel_colormap, marker='o', vmin=vmin, vmax=vmax)
        cbar = plt.colorbar(scatter, label='Peak reflectivity [dB]')

        cbar_ticks = np.arange(vmin, vmax + z_step, z_step)
        cbar.set_ticks(cbar_ticks)
        cbar.ax.minorticks_on()
        cbar.ax.yaxis.set_minor_locator(AutoMinorLocator(minor_tick_num))  # For vertical colorbars

        ax.yaxis.set_major_formatter(axis_formatter_degree)

        plt.xlabel('Soil moisture $[\mathrm{m}^3 \mathrm{m}^{-3}]$')
        plt.xlim([0.0, 0.4])
        x_ticks = np.arange(0, 0.6, 0.1)
        ax.set_xticks(x_ticks)
        ax.xaxis.set_minor_locator(AutoMinorLocator(2))

        # Determine the tick range for y-axis
        y_ticks = np.arange(0, 70, 10)
        ax.set_yticks(y_ticks)
        ax.set_ylim([0, 65])
        ax.yaxis.set_minor_locator(AutoMinorLocator(2))

        plt.ylabel('$\\theta_\mathrm{i}$')
        plt.tight_layout()
        plt.tight_layout()
        img_name = img_tag + f'_peak_refl_vs_sm_inc_angle_alt_{alt_val}_all_az'
        save_figure(fig, folder_path, img_name, fig_save_types=fig_save_types)
        plt.close(fig)


def save_figure(fig, fig_out_folder, img_save_name, tf_save_fig=True, fig_save_types: Optional[list[str]] = None):
    """
    save figure

    :param fig: figure object
    :param fig_out_folder: image saving folder
    :param img_save_name: image save name
    :param tf_save_fig: save the figure?
    :param fig_save_types: type of image? png, eps, pdf, svg ...
    :return: True if image is saved, else False
    """
    if not tf_save_fig:
        return False
    if fig_save_types is None:
        fig_save_types = default_save_type
    elif type(fig_save_types) is str:
        fig_save_types = [fig_save_types]
    for fig_type in fig_save_types:
        name = f"{img_save_name.split('.')[0]}.{fig_type}"
        fig.savefig(os.path.join(fig_out_folder, name), format=fig_type, bbox_inches='tight')
    return True


from time import sleep
from typing import Optional
import os
import requests
import shutil
import zipfile


ANT_PTRN_URL = 'https://bitbucket.org/usc_mixil/cygnss_antenna_patterns/get/master.zip'
MAX_CONNECTIONS_TRIES = 5
CHUNK_SIZE = 1024 * 1024  # 1 MB


def download_file(file_url: str, output_folder: str) -> Optional[str]:
    """
    download the file with url into folder output_folder

    :param file_url: url of the file
    :type file_url: str
    :param output_folder: saving folder
    :type output_folder: str
    :param auth: username or pass
    :type auth: tuple of str or None
    :param url_md5_checksum: md5 checksum of the downloaded file, if None, the code will find it
    :type url_md5_checksum: str
    :return: downloaded file path
    :rtype: str
    """
    num_redownload = 0
    file_name = file_url.split('/')[-1]
    out_file = os.path.join(output_folder, file_name)
    out_file_temp = f'{out_file:s}.incomplete'
    for i_try in range(MAX_CONNECTIONS_TRIES):
        try:
            with requests.get(file_url, stream=True) as response:
                if response.status_code == 404:
                    return None
                elif response.status_code == 401:  # Auth Error
                    raise requests.exceptions.RequestException
                response.raise_for_status()
                response.raw.decode_content = True
                file_size = int(response.headers.get('content-length', 0))
                free_disk_space = shutil.disk_usage(output_folder).free
                if file_size > free_disk_space:
                    def bytes2mb(x: float) -> int: return int(x / 1024 / 1024)
                    raise IOError(f'No enough space in the disk. file size: {bytes2mb(file_size):d} MB, free space {bytes2mb(free_disk_space):d}')
                with open(out_file_temp, "wb") as target_file:
                    for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                        target_file.write(chunk)
        except (requests.exceptions.ConnectionError, requests.exceptions.HTTPError):
            if i_try < MAX_CONNECTIONS_TRIES:
                print('Connection Error sleep for 5 sec')
                sleep(5)
            else:
                raise RuntimeError(f'a connection error was raised when trying to access {file_url:s}')
        else:  # if it was successful
            shutil.move(out_file_temp, out_file)
            break
    return out_file

def download_cyg_antenna_patterns(antenna_patterns_folder_path=None):
    if antenna_patterns_folder_path is None:
        antenna_patterns_folder_path = os.path.join(os.environ["CYGNSS_PATH"], 'cygnss_antenna_patterns')
    if not os.path.isdir(antenna_patterns_folder_path):
        os.makedirs(antenna_patterns_folder_path)

    files_list = os.listdir(antenna_patterns_folder_path)
    if 'antennaRx_CYGNSS_Obs1_Nadir02_Starboard_processed' in files_list and 'antennaRx_CYGNSS_Obs1_Nadir01_Port_processed' in files_list:
        ant_ptrn_files_path = [os.path.join(antenna_patterns_folder_path, 'antennaRx_CYGNSS_Obs1_Nadir02_Starboard_processed'),
                               os.path.join(antenna_patterns_folder_path, 'antennaRx_CYGNSS_Obs1_Nadir01_Port_processed')]
        return ant_ptrn_files_path

    ant_ptrn_zip_path = os.path.join(antenna_patterns_folder_path, ANT_PTRN_URL.split('/')[-1])
    if not os.path.exists(ant_ptrn_zip_path):
        ant_ptrn_zip_path = download_file(ANT_PTRN_URL, antenna_patterns_folder_path)

    ant_ptrn_files_path = extract_ant_ptrn_zip(ant_ptrn_zip_path, antenna_patterns_folder_path)
    os.remove(ant_ptrn_zip_path)  # remove zip file
    return ant_ptrn_files_path


def extract_ant_ptrn_zip(zip_file_path, out_path):

    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:

        files_list = zip_ref.namelist()
        extracted_files = list()
        for file_name in files_list:
            if file_name.endswith('processed'):
                zip_ref.extract(file_name, path=out_path)
                shutil.move(os.path.join(out_path, file_name), os.path.join(out_path, file_name.split(os.sep)[-1]))
                extracted_files.append(os.path.join(out_path, file_name.split(os.sep)[-1]))

    os.rmdir(os.path.join(out_path, file_name.split(os.sep)[0]))
    return extracted_files


if __name__ == '__main__':
    download_cyg_antenna_patterns()




#!/usr/bin/env python3
import os
import datetime
import datetime as dt
from typing import Union, Optional
from cygnsslib.cyg import land_flags_check
from osgeo import gdal, ogr
from cygnsslib.data_downloader.download_cygnss import get_cyg_file, download_cyg_files
from geographiclib.geodesic import Geodesic
from multiprocessing import Pool
from netCDF4 import Dataset
from tqdm import tqdm
import numpy as np
import pandas as pd


gdal.UseExceptions()


def roll_lon(val):
    if isinstance(val, np.ndarray):
        val[val > 180.0] += -360.0
    else:
        if val > 180.0:
            val += -360.0
    return val


def find_var_in_day(cygnss_l1_path: str, i_date: datetime.date, download_cygnss_data: bool, thresh_ddm_snr: float, ddm_quality_filter: int,
                    bbox: list, poly: ogr.Geometry, tf_poly: bool, lon_mode360: bool, ref_pos: Union[list, np.ndarray], radius: float, var_names,
                    tf_print_screen: bool, pbar: Optional[tqdm] = None):
    """
    Internal function for extracting CYGNSS variables of a specific day

    :param cygnss_l1_path: path of the main folder of CYGNSS L1 data (after the versions)
    :param i_date: date
    :param download_cygnss_data: download CYGNSS L1 data if they're not available?
    :param thresh_ddm_snr: DDM SNR threshold, above this value the DDM map image will be saved
    :param ddm_quality_filter: our quality flog number, above this number is low
    :param bbox:
    :param poly:
    :param tf_poly:
    :param lon_mode360:
    :param ref_pos:
    :param radius:
    :param var_names:
    :param tf_print_screen:
    :param pbar:
    :return:
    """
    geod = Geodesic.WGS84
    day_pnt_list_high = list()
    day_pnt_list_low = list()
    day = i_date.timetuple().tm_yday
    cyg_day_folder = os.path.join(cygnss_l1_path, f'{i_date.year:04d}', f'{day:03d}')
    for sc_num in np.arange(1, 9):
        if pbar is not None:
            pbar.update()
        filename = get_cyg_file(cyg_day_folder, sc_num)
        if filename is None and download_cygnss_data:
            file_name = download_cyg_files(i_date.year, day, sc_num, cygnss_l1_path=cygnss_l1_path)
            filename = None if (not file_name) else file_name[0]

        if filename is None:
            continue
        if tf_print_screen:
            print(filename)
        file_path = os.path.join(cyg_day_folder, filename)
        try:
            nc_file = Dataset(file_path)
            nc_file.set_auto_maskandscale(False)
            tsc = nc_file.time_coverage_start
            ddm_timestamp_utc = nc_file.variables["ddm_timestamp_utc"]
            sample = nc_file.variables["sample"]
            sc_num = nc_file.variables["spacecraft_num"]
            sp_lat = np.array(nc_file.variables["sp_lat"])
            sp_lon = np.array(nc_file.variables["sp_lon"])
            sp_inc_angle = nc_file.variables["sp_inc_angle"]
            ddm_snr = nc_file.variables["ddm_snr"]

        except (OSError, RuntimeError) as e:
            print(e)
            raise RuntimeError(f'Mostly, the file is damaged, try to re-download it again \n file path {file_path}')
        sp_lon_rolled = sp_lon[:]
        tf_rolled = sp_lon[:] > 180.0
        sp_lon_rolled[sp_lon[:] > 180.0] -= 360.0
        tf_in_box = (sp_lat[:] > bbox[2]) & (sp_lat[:] < bbox[3]) & (sp_lon_rolled > bbox[0]) & (sp_lon_rolled < bbox[1])
        tf_chan = np.sum(tf_in_box, axis=0) != 0
        for i_chan, tf_hit in enumerate(tf_chan):
            if tf_hit:
                tf_sel = tf_in_box[:, i_chan]
                sample_sel = sample[tf_sel]
                for i_samp in sample_sel:
                    if tf_poly:
                        pt_cur = ogr.Geometry(ogr.wkbPoint)
                        sel_sp_lon = sp_lon[i_samp, i_chan]
                        if lon_mode360 and sel_sp_lon < 0:
                            sel_sp_lon += 360.0
                        pt_cur.AddPoint(float(sel_sp_lon), float(sp_lat[i_samp, i_chan]))
                        sp_is_in_poly = pt_cur.Within(poly)
                        pt_cur = None
                    else:
                        g_dist = geod.Inverse(ref_pos[0], ref_pos[1], float(sp_lat[i_samp, i_chan]), float(sp_lon_rolled[i_samp, i_chan]))["s12"]
                        sp_is_in_poly = g_dist < radius
                        if sp_is_in_poly and tf_print_screen:
                            print(f'(dist={g_dist:2.0f} m)(ddm_snr={ddm_snr[i_samp, i_chan]:2.1f} dB)')

                    if sp_is_in_poly:
                        qflags = np.array(nc_file.variables["quality_flags"][i_samp, i_chan])
                        sp_delay_row = int(nc_file.variables['brcs_ddm_peak_bin_delay_row'][i_samp, i_chan])
                        sp_dopp_col = int(nc_file.variables['brcs_ddm_peak_bin_dopp_col'][i_samp, i_chan])
                        qflags_msg_list, land_flag, our_flags = land_flags_check(qflags, nc_file.variables["sp_rx_gain"][i_samp, i_chan],
                                                                                 sp_delay_row, sp_dopp_col)

                        ddm_data = {var_: nc_file.variables[var_][i_samp, i_chan] for var_ in var_names}
                        timestamp_utc = np.timedelta64(int(ddm_timestamp_utc[i_samp] * 1e9), 'ns') + np.datetime64(tsc[:-1])
                        timestamp_utc_str = np.datetime_as_string(timestamp_utc)
                        ddm_metadata = {'ddm_timestamp_utc_str': timestamp_utc_str,
                                        'year': i_date.year,
                                        'day': day,
                                        'spacecraft_num': int(sc_num[0]),
                                        'channel': int(i_chan + 1),
                                        'sample_zero_based': int(i_samp),
                                        'sp_lat': float(sp_lat[i_samp, i_chan]),
                                        'sp_lon': float(sp_lon_rolled[i_samp, i_chan]),
                                        'sp_inc_angle': float(sp_inc_angle[i_samp, i_chan]),
                                        'ddm_snr': float(ddm_snr[i_samp, i_chan]),
                                        'DDM_quality': our_flags,
                                        'is_over_land': bool(land_flag),
                                        'quality_flags_msg': ', '.join(qflags_msg_list)}
                        ddm_data.update(ddm_metadata)
                        if ddm_snr[i_samp, i_chan] < thresh_ddm_snr or our_flags > ddm_quality_filter:
                            day_pnt_list_low.append(ddm_data)
                        else:
                            day_pnt_list_high.append(ddm_data)
        nc_file.close()
    return day_pnt_list_high, day_pnt_list_low


def _get_var_stat_from_poly_or_circle(date_list: list[dt.date], poly, ref_pos: Union[tuple[float], list[float]], radius: float,
                                      tf_poly: bool, var_list: list[str], thresh_ddm_snr: float, out_root: str, download_cygnss_data: bool,
                                      cygnss_l1_path: Optional[str], out_options: Optional[dict[str]]):

    if out_options is None:
        out_options = dict()
    plt_tag = out_options["plt_tag"] if ("plt_tag" in out_options) else ""
    tf_print_screen = not out_options['silent_mode'] if ("silent_mode" in out_options) else True
    pbar_tf = True if not tf_print_screen else False
    pbar_tf = out_options['progress_bar'] if ("progress_bar" in out_options) else pbar_tf
    parallel = out_options['multithreads'] if ('multithreads' in out_options) else False
    save_below_thresh = out_options['save_data_below_threshold'] if ('save_data_below_threshold' in out_options) else True
    ddm_quality_filter = out_options['ddm_quality_filter'] if ('ddm_quality_filter' in out_options) else 3
    sheet_save_type = 'xlsx'
    if 'sheet_type' in out_options:
        if out_options['sheet_type'].lower() == 'csv':
            sheet_save_type = 'csv'

    if cygnss_l1_path is None:
        cygnss_l1_path = os.environ.get("CYGNSS_L1_PATH")
        if cygnss_l1_path is None:
            raise ValueError("$CYGNSS_L1_PATH environment variable need to be set, or use cygnss_l1_path input parameter")

    # Define bounding box containing circle of input radius for efficient data exclusion
    geod = Geodesic.WGS84
    if tf_poly:
        bbox = list(poly.GetEnvelope())
        ref_pos = [0.0, 0.0]
        ref_pos[1], ref_pos[0], *_ = poly.Centroid().GetPoint()
        ref_pos[1] = roll_lon(ref_pos[1])

        radius = 10
    else:
        g_n = geod.Direct(ref_pos[0], ref_pos[1], 0, radius)
        g_e = geod.Direct(ref_pos[0], ref_pos[1], 90, radius)
        g_s = geod.Direct(ref_pos[0], ref_pos[1], 180, radius)
        g_w = geod.Direct(ref_pos[0], ref_pos[1], 270, radius)
        bbox = [g_w["lon2"], g_e["lon2"], g_s["lat2"], g_n["lat2"]]
    lon_mode360 = False
    if bbox[0] > 180.0 or bbox[1] > 180.0:
        lon_mode360 = True
    bbox[0] = roll_lon(bbox[0])
    bbox[1] = roll_lon(bbox[1])
    var_stats_above_high = {var_: [0, np.inf, -np.inf, 0.0] for var_ in var_list}  # Num, min, max, mean
    var_stats_above_low = {var_: [0, np.inf, -np.inf, 0.0] for var_ in var_list}
    file_tag = os.path.basename(out_root)
    if parallel:
        # date_list_ = tqdm(date_list, total=len(date_list), desc='Searching CYGNSS files', unit='day') if pbar_tf else date_list
        pbar = None
        gdal.UseExceptions()

        with Pool() as pool:
            args = ((cygnss_l1_path, i_date, download_cygnss_data, thresh_ddm_snr, ddm_quality_filter, bbox, poly, tf_poly, lon_mode360,
                     ref_pos, radius, var_list, tf_print_screen, pbar) for i_date in date_list)
            results = [pool.apply_async(find_var_in_day, args=arg) for arg in args]
            if pbar_tf:
                pool_loop = tqdm(enumerate(results), desc=f'Stats: {file_tag}', unit='day', total=len(date_list))
            else:
                pool_loop = enumerate(results)
            for iresult, r in pool_loop:
                day_pnt_list_high, day_pnt_list_low = r.get()
                for point_list, var_stats in zip([day_pnt_list_high, day_pnt_list_low], [var_stats_above_high, var_stats_above_low]):
                    if point_list:
                        for var_name in var_list:
                            var_stats[var_name][0] += len(point_list)  # num
                            var_vals = np.array([ddm_data[var_name] for ddm_data in point_list])
                            min_val = np.min(var_vals)
                            max_val = np.max(var_vals)
                            sum_ = np.sum(var_vals)
                            var_stats[var_name][1] = np.minimum(var_stats[var_name][1], min_val)
                            var_stats[var_name][2] = np.maximum(var_stats[var_name][2], max_val)
                            var_stats[var_name][3] += sum_
        del args, day_pnt_list_high, day_pnt_list_low, pool_loop
    else:
        pbar = tqdm(total=len(date_list) * 8, desc=f'Stats: {file_tag}', unit='file') if pbar_tf else None
        for i_date in date_list:
            day_pnt_list_high, day_pnt_list_low = find_var_in_day(cygnss_l1_path, i_date, download_cygnss_data, thresh_ddm_snr, ddm_quality_filter,
                                                                  bbox, poly, tf_poly, lon_mode360, ref_pos, radius, var_list, tf_print_screen,
                                                                  pbar)
            for point_list, var_stats in zip([day_pnt_list_high, day_pnt_list_low], [var_stats_above_high, var_stats_above_low]):
                if point_list:
                    for var_name in var_list:
                        var_stats[var_name][0] += len(point_list)  # num
                        var_vals = np.array([ddm_data[var_name] for ddm_data in point_list])
                        min_val = np.min(var_vals)
                        max_val = np.max(var_vals)
                        sum_ = np.sum(var_vals)
                        var_stats[var_name][1] = np.minimum(var_stats[var_name][1], min_val)
                        var_stats[var_name][2] = np.maximum(var_stats[var_name][2], max_val)
                        var_stats[var_name][3] += sum_
    if pbar is not None:
        pbar.close()

    # Get average
    for var_stats in [var_stats_above_high, var_stats_above_low]:
        for var_name in var_list:
            if var_stats[var_name][0] > 0:
                var_stats[var_name][3] = var_stats[var_name][3] / var_stats[var_name][0]

    out_high_fp = out_root + '_high.csv'
    out_low_fp = out_root + '_low.csv'
    if not os.path.isdir(os.path.dirname(out_high_fp)):
        os.makedirs(os.path.dirname(out_high_fp))
    df_high = pd.DataFrame(var_stats_above_high, index=['Num', 'Min', 'Max', 'Mean']).T
    df_high.to_csv(out_high_fp)
    print(f'Stats high stave at {out_high_fp}')
    if save_below_thresh:
        df_low = pd.DataFrame(var_stats_above_low, index=['Num', 'Min', 'Max', 'Mean']).T
        df_low.to_csv(out_low_fp)
        print(f'Stats low stave at {out_low_fp}')


def get_var_stat_within_box_between_dates(st_date: dt.date, end_date: dt.date, lat_lim: Union[list[float], np.ndarray],
                                          lon_lim: Union[list[float], np.ndarray], var_list, out_folder_path: str, out_file_tag: str,
                                          thresh_ddm_snr: float = -9999.0, download_cygnss_data: bool = True,
                                          out_options: Optional[dict] = None, cygnss_l1_dir: Optional[str] = None):
    geo_shape = ogr.Geometry(ogr.wkbLinearRing)
    lat_lim = np.sort(lat_lim)
    lon_lim = np.sort(lon_lim)

    geo_shape.AddPoint(lon_lim[0], lat_lim[0])
    geo_shape.AddPoint(lon_lim[1], lat_lim[0])
    geo_shape.AddPoint(lon_lim[1], lat_lim[1])
    geo_shape.AddPoint(lon_lim[0], lat_lim[1])
    geo_shape.AddPoint(lon_lim[0], lat_lim[0])

    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(geo_shape)
    ref_pos = [0.0, 0.0]
    ref_pos[1], ref_pos[0], *_ = poly.Centroid().GetPoint()
    ref_pos[1] = roll_lon(ref_pos[1])
    radius = 10
    tf_poly = True
    geo_shape = None
    num_days = (end_date - st_date).days + 1
    date_list = [st_date + dt.timedelta(days=iday) for iday in range(0, num_days)]
    out_root = os.path.join(out_folder_path, out_file_tag)
    return _get_var_stat_from_poly_or_circle(date_list, poly, ref_pos, radius, tf_poly, var_list, thresh_ddm_snr, out_root, download_cygnss_data,
                                             cygnss_l1_dir, out_options)


def get_var_stat_within_radius(year: int, daylist: Union[list[int], np.ndarray, int], ref_pos: tuple[float, float], radius: float, var_list,
                               out_folder_path: str, out_file_tag: str, thresh_ddm_snr: float = -9999.0, download_cygnss_data: bool = True,
                               out_options: Optional[dict] = None, cygnss_l1_dir: Optional[str] = None):

    geod = Geodesic.WGS84
    geo_shape = ogr.Geometry(ogr.wkbLinearRing)

    for angle in np.arange(0, 361, 1):
        poly_point = geod.Direct(ref_pos[0], ref_pos[1], angle, radius)
        geo_shape.AddPoint(poly_point["lon2"], poly_point["lat2"])

    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(geo_shape)
    tf_poly = False
    geo_shape = None

    year = int(year)
    if np.size(daylist) == 1:
        daylist = [int(daylist)]
    date_list = [dt.date(year, month=1, day=1) + dt.timedelta(days=int(iday) - 1) for iday in daylist]
    out_root = os.path.join(out_folder_path, out_file_tag)
    return _get_var_stat_from_poly_or_circle(date_list, poly, ref_pos, radius, tf_poly, var_list, thresh_ddm_snr, out_root, download_cygnss_data,
                                             cygnss_l1_dir, out_options)


def get_var_stat_within_box(year: int, daylist: Union[list[int], np.ndarray, int], lat_lim: Union[list[float], np.ndarray],
                            lon_lim: Union[list[float], np.ndarray], var_list, out_folder_path: str, out_file_tag: str, thresh_ddm_snr: float = -9999.0,
                            download_cygnss_data: bool = True, out_options: Optional[dict] = None, cygnss_l1_dir: Optional[str] = None):
    geo_shape = ogr.Geometry(ogr.wkbLinearRing)
    lat_lim = np.sort(lat_lim)
    lon_lim = np.sort(lon_lim)

    geo_shape.AddPoint(lon_lim[0], lat_lim[0])
    geo_shape.AddPoint(lon_lim[1], lat_lim[0])
    geo_shape.AddPoint(lon_lim[1], lat_lim[1])
    geo_shape.AddPoint(lon_lim[0], lat_lim[1])
    geo_shape.AddPoint(lon_lim[0], lat_lim[0])

    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(geo_shape)
    ref_pos = [0.0, 0.0]
    ref_pos[1], ref_pos[0], *_ = poly.Centroid().GetPoint()
    ref_pos[1] = roll_lon(ref_pos[1])
    radius = 10
    tf_poly = True
    geo_shape = None

    year = int(year)
    if np.size(daylist) == 1:
        daylist = [int(daylist)]
    date_list = [dt.date(year, month=1, day=1) + dt.timedelta(days=int(iday) - 1) for iday in daylist]
    out_root = os.path.join(out_folder_path, out_file_tag)
    return _get_var_stat_from_poly_or_circle(date_list, poly, ref_pos, radius, tf_poly, var_list, thresh_ddm_snr, out_root, download_cygnss_data,
                                             cygnss_l1_dir, out_options)


def get_var_stat_within_radius_between_dates(st_date: dt.date, end_date: dt.date, ref_pos: tuple[float, float], radius: float, var_list: list[str],
                                             out_folder_path: str, out_file_tag: str, thresh_ddm_snr: float = -9999.0,
                                             download_cygnss_data: bool = True, out_options: Optional[dict] = None,
                                             cygnss_l1_dir: Optional[str] = None):
    geod = Geodesic.WGS84
    geo_shape = ogr.Geometry(ogr.wkbLinearRing)

    for angle in np.arange(0, 361, 1):
        poly_point = geod.Direct(ref_pos[0], ref_pos[1], angle, radius)
        geo_shape.AddPoint(poly_point["lon2"], poly_point["lat2"])

    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(geo_shape)
    tf_poly = False
    geo_shape = None

    num_days = (end_date - st_date).days + 1
    date_list = [st_date + dt.timedelta(days=iday) for iday in range(0, num_days)]
    out_root = os.path.join(out_folder_path, out_file_tag)

    return _get_var_stat_from_poly_or_circle(date_list, poly, ref_pos, radius, tf_poly, var_list, thresh_ddm_snr, out_root, download_cygnss_data,
                                             cygnss_l1_dir, out_options)


if __name__ == '__main__':
    lat = 32.5187357
    lon = -106.7434509
    margin = 0.2
    start_time = dt.date(2021, 1, 1)
    end_time = dt.date(2022, 1, 1)
    lat_ = [lat - margin, lat+margin]
    lon_ = [lon - margin, lon+margin]
    param_list = ['sp_inc_angle', 'tx_to_sp_range', 'rx_to_sp_range']
    get_var_stat_within_box_between_dates(start_time, end_time, lat_, lon_, param_list, 'test_output', 'z1_data',
                                          3, False)



import argparse
import json
import os
from collections import defaultdict
from cygnsslib.plotting import (plot_varying_sp_inc_angle, plot_varying_sc_az_angle, plot_reflectivity_x_az_lg_inc_angle,
                                plot_reflectivity_x_az_lg_sm, plot_reflectivity_vs_inc_angle_sm,
                                plot_reflectivity_vs_inc_angle_sm_all_az, plot_reflectivity_x_sm_lg_inc_angle_all_az)


# Function to load JSON data from a file
def load_json(file_path):
    with open(file_path, 'r') as file:
        data = json.load(file)
    return data

# Function to group data by 'smap_sm', 'sc_alt', 'sp_inc_angle', and 'sc_az_angle' values
def group_data(data):
    grouped_data = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list))))
    for _, item in data.items():
        smap_sm = item.get('smap_sm')
        sc_alt = item.get('sc_alt')
        sp_inc_angle = item.get('sp_inc_angle')
        sc_az_angle = item.get('sc_az_angle')

        if smap_sm is not None and sc_alt is not None and sp_inc_angle is not None and sc_az_angle is not None:
            # Determine the smap_sm group by 0.05 ranges
            smap_sm_group = int(smap_sm // 0.05) * 0.05
            # Determine the sc_alt group by 10000 ranges
            sc_alt_group = int(sc_alt // 10000) * 10000
            # Determine the sp_inc_angle group by 5 ranges
            sp_inc_angle_group = int(sp_inc_angle // 5) * 5
            # Determine the sc_az_angle group by 5 ranges
            sc_az_angle_group = int(sc_az_angle // 5) * 5

            grouped_data[smap_sm_group][sc_alt_group][sp_inc_angle_group][sc_az_angle_group].append(item)

    return grouped_data



def analyze_cyg(json_fp, fig_save_types=None, db_lim=None, plt_options=None, separate_legend=False):
    default_plt_options = ['timeseries_az', 'timeseries_inc', 'inc_vs_sm_all_az', 'inc_vs_sm', 'az_lg_sm', 'az_lg_inc', 'sm_lg_inc']

    if plt_options is None:
        plt_options = default_plt_options
    else:
        for plt_opt in plt_options:
            if plt_opt not in default_plt_options:
                print(f'Not supported plot option: {plt_opt}, supported options: '+ ','.join(default_plt_options))

    if fig_save_types is None:
        fig_save_types = ['png']
    fn_split = os.path.basename(json_fp).split('_')
    _tag_idx = [idx for idx, txt in enumerate(fn_split) if txt.isnumeric()]
    if not _tag_idx:
        _tag_idx = 2
    else:
        _tag_idx = _tag_idx[0]
    img_tag = '_'.join(fn_split[:(_tag_idx+1)])

    data = load_json(json_fp)
    img_folder_name = f'{img_tag}_plots'
    folder_path = os.path.join(os.path.dirname(json_fp), img_folder_name)
    if not os.path.isdir(folder_path):
        os.makedirs(folder_path)
    grouped_data = group_data(data)
    if 'timeseries_inc' in plt_options:
        plot_varying_sp_inc_angle(grouped_data, folder_path, img_tag, fig_save_types)
    if 'timeseries_az' in plt_options:
        plot_varying_sc_az_angle(grouped_data, folder_path, img_tag, fig_save_types)
    if 'inc_vs_sm_all_az' in plt_options:
        plot_reflectivity_vs_inc_angle_sm_all_az(grouped_data, folder_path, img_tag, fig_save_types)
    if 'inc_vs_sm' in plt_options:
        plot_reflectivity_vs_inc_angle_sm(grouped_data, folder_path, img_tag, fig_save_types)
    if 'az_lg_sm' in plt_options:
        plot_reflectivity_x_az_lg_sm(grouped_data, folder_path, img_tag, fig_save_types, db_lim)
    if 'az_lg_inc' in plt_options:
        plot_reflectivity_x_az_lg_inc_angle(grouped_data, folder_path, img_tag, fig_save_types, db_lim, separate_legend)
    if 'sm_lg_inc' in plt_options:
        plot_reflectivity_x_sm_lg_inc_angle_all_az(grouped_data, folder_path, img_tag, fig_save_types, db_lim, separate_legend)


if __name__ == '__main__':
    opt_desc = ('1. timeseries_az: Plot timeseries of peak reflectivity, az_angle in legend \n '
                '2. timeseries_inc: Plot timeseries of peak reflectivity, incidence angle in legend \n '
                '3. inc_vs_sm_all_az: 2D plot of peak reflectivity, x-asis is soil moisture, y-axis incidence angle, all az angles in same plot \n '
                '4. inc_vs_sm: 2D plot of peak reflectivity, x-asis is soil moisture, y-axis incidence angle \n '
                '5. az_lg_sm: Plot peak reflectivity, x-asis azimuth angle, legend is soil moisture group \n '
                '6. az_lg_inc: Plot peak reflectivity, x-asis azimuth angle, legend is incidence angle group \n'
                '7. sm_lg_inc: Plot peak reflectivity, x-asis soil moisture, legend is incidence angle group, for all az angles')
    parser = argparse.ArgumentParser(description=f'Plot CYGNSS reflectivity: --plot_option \n {opt_desc}')
    parser.add_argument('input_path', type=str, help='Path to JSON file or folder has json files')
    parser.add_argument('--img_save_type', nargs='+', default=None, type=str, help='Image save types, i.e. png svg pdf')
    parser.add_argument('--dblim', nargs='+', default=None, type=int, help='dB lower and upper limit')
    parser.add_argument('--plot_option', nargs='+', default=None, type=str, help='Plotting options: timeseries_az, timeseries_inc, inc_vs_sm_all_az, inc_vs_sm, az_lg_sm, az_lg_inc, sm_lg_inc')
    parser.add_argument('--separate_legend', action='store_true', default=False, help='Plot legend in separate figure')

    args = parser.parse_args()
    if os.path.isdir(args.input_path):
        fps = [f.path for f in os.scandir(args.input_path) if f.name.endswith('.json')]
    else:
        fps = [args.input_path]
    for fp in fps:
        print(fp)
        analyze_cyg(fp, args.img_save_type, args.dblim, args.plot_option, args.separate_legend)



import datetime as dt
import fnmatch
from typing import Optional, Union
import numpy as np
import os
import warnings
from cygnsslib.data_downloader.download_cygnss import download_cyg_files, download_cyg_files_between_date


def download_rawif_cyg_files_between_date(st_date: dt.date, end_date: dt.date, list_sc_num: Optional[Union[list[int], int]] = None,
                                          cyg_data_ver: Optional[str] = None, cygnss_l1_path: Optional[str] = None,
                                          checksum_exist_file: bool = False, force_download: bool = False, download_l1_data: bool = False):
    """
    download RAWIF CYGNSS data between two dates (including start and end date)

    :param st_date: start date
    :param end_date: end date
    :param list_sc_num: list of cygnss spacecraft numbers (1-8), if None will download all SCs
    :param cygnss_l1_path: path of the cygnss L1 data (default: os.environ.get('CYGNSS_L1_PATH')), see description for more details
    :param force_download: re-download the file if it exist?
    :param download_l1_data: when there is a Rawif data, download its L1 data with it, env $CYGNSS_L1_PATH var should point to the folder of L1 data.
    :param cyg_data_ver: CYGNSS data L1 data version
    :param checksum_exist_file: apply check sum to existing files
    :return:
    """
    if cygnss_l1_path is None:
        cygnss_l1_path = os.environ.get('CYGNSS_L1_PATH')

    # check if the folder name is not raw_if, if not, change the folder name
    folder_list = cygnss_l1_path.split(os.path.sep)
    if not folder_list[-1]:
        folder_list.pop(-1)
    if 'raw_if' not in folder_list[-1]:
        folder_list[-1] = 'raw_if'
    cyg_rawif_path = os.sep.join(folder_list)

    cyg_data_lvl = 'RAW'
    download_cyg_files_between_date(st_date, end_date, list_sc_num, cyg_data_ver, cyg_data_lvl, cyg_rawif_path, checksum_exist_file, force_download)
    if download_l1_data:
        cyg_data_lvl = 'L1'
        download_cyg_files_between_date(st_date, end_date, list_sc_num, cyg_data_ver, cyg_data_lvl, cygnss_l1_path, checksum_exist_file, force_download)


def download_cyg_rawif_files(data_year: int, list_data_day: Union[list[int], int], list_sc_num: Optional[Union[list[int], int]] = None,
                             cyg_data_ver: Optional[str] = None, cygnss_l1_path: Optional[str] = None, checksum_exist_file: bool = False,
                             force_download: bool = False, download_l1_data: bool = False):
    """

    download the raw_if cygnss data,
    if cygnss_l1_path or os.environ.get('CYGNSS_L1_PATH') point to a folder with name not "raw_if", it will save the data in a raw_if folder in the
    parent dir.

    :param data_year: list of data years
    :param list_data_day: list of data days
    :param list_sc_num: list of cygnss spacecraft numbers (1-8), if None will download all SCs
    :param cygnss_l1_path: path of the cygnss L1 data (default: os.environ.get('CYGNSS_L1_PATH')), see description for more details
    :param force_download: re-download the file if it exist?
    :param download_l1_data: when there is a Rawif data, download its L1 data with it, env $CYGNSS_L1_PATH var should point to the folder of L1 data.
    :param cyg_data_ver: CYGNSS data L1 data version
    :param checksum_exist_file: apply check sum to existing files
    :return:
    """
    if cygnss_l1_path is None:
        cygnss_l1_path = os.environ.get('CYGNSS_L1_PATH')

    # check if the folder name is not raw_if, if not, change the folder name
    folder_list = cygnss_l1_path.split(os.path.sep)
    if not folder_list[-1]:
        folder_list.pop(-1)
    if 'raw_if' not in folder_list[-1]:
        folder_list[-1] = 'raw_if'
    cyg_rawif_path = os.sep.join(folder_list)

    cyg_data_lvl = 'raw'
    download_cyg_files(data_year, list_data_day, list_sc_num, cyg_data_ver, cyg_data_lvl, cyg_rawif_path, checksum_exist_file, force_download)

    if download_l1_data:
        cyg_data_lvl = 'L1'
        download_cyg_files(data_year, list_data_day, list_sc_num, cyg_data_ver, cyg_data_lvl, cygnss_l1_path, checksum_exist_file, force_download)


def get_cyg_rawif_files(cyg_day_folder, sc_num):
    """
    check if the file exist and return the file name, if not exist return None.
    if exist it will return list of the files

    :param cyg_day_folder: cygnss day folder
    :type cyg_day_folder: str
    :param sc_num: spacecraft number
    :type sc_num: int
    :return: file name
    :rtype: str
    """
    _files_flag = np.zeros(2).astype(bool)
    result = []
    pattern = "cyg{:02d}*.bin".format(sc_num)
    for root, dirs, files in os.walk(cyg_day_folder):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(name)

    if len(result) == 0:
        return None
    else:
        files_name_list = list()
        for file_name in result:
            if 'data' in file_name:
                _files_flag[0] = True
                files_name_list.append(file_name)
            elif 'meta' in file_name:
                _files_flag[1] = True
                files_name_list.append(file_name)

    if not _files_flag.all():
        warnings.warn(f"couldn't find both data and the metadata files in {cyg_day_folder:s}, sc: {sc_num:d}, try to download both",
                      RuntimeWarning)
        return None

    return files_name_list


if __name__ == '__main__':
    down_start_date = dt.date(year=2020, month=8, day=4)
    down_end_date = dt.date(year=2020, month=8, day=4)
    download_rawif_cyg_files_between_date(down_start_date, down_end_date)
    download_cyg_rawif_files(data_year=2020, list_data_day=227)
